"""The decision module handles the three planning levels

Currently, only pre-specified planning is implemented.

The choices made in the three planning levels influence the set of interventions
and assets available within a model run.

The interventions available in a model run are stored in the
:class:`~smif.intervention.InterventionRegister`.

When pre-specified planning are declared, each of the corresponding
interventions in the InterventionRegister are moved to the BuiltInterventionRegister.

Once pre-specified planning is instantiated, the action space for rule-based and
optimisation approaches can be generated from the remaining Interventions in the
InterventionRegister.

"""
__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


from abc import ABCMeta, abstractmethod


class DecisionManager(object):
    """A DecisionManager is initialised with one or more model run strategies that refer to
    DecisionModules such as pre-specified planning, a rule-based models or multi-objective
    optimisation. These implementations influence the combination and ordering of decision
    iterations and model timesteps that need to be performed by the model runner.

    The DecisionManager presents a simple decision loop interface to the model runner, in the
    form of a generator which allows the model runner to iterate over the collection of
    independent simulations required at each step.

    (Not yet implemented.) A DecisionManager collates the output of the decision algorithms and
    writes the post-decision state through a DataHandle. This allows Models to access a given
    decision state (identified uniquely by timestep and decision iteration id).

    (Not yet implemented.) A DecisionManager may also pass a DataHandle down to a
    DecisionModule, allowing the DecisionModule to access model results from previous timesteps
    and decision iterations when making decisions.

    Arguments
    ---------
    timesteps: list
    strategies: list
    """

    def __init__(self, timesteps, strategies):
        self._timesteps = timesteps
        self._strategies = strategies
        self._decision_modules = []

        self._set_up_decision_modules()

    def _set_up_decision_modules(self):
        interventions = []

        for strategy in self._strategies:
            if strategy['strategy'] == 'pre-specified-planning':
                interventions.extend(strategy['interventions'])
            else:
                msg = "Only pre-specified planning strategies are implemented"
                raise NotImplementedError(msg)

        self._decision_modules = [PreSpecified(self._timesteps, interventions)]

    def decision_loop(self):
        """Generate bundles of simulation steps to run.

        Each iteration returns a dict: {decision_iteration (int) => list of timesteps (int)}

        With only pre-specified planning, there is a single step in the loop, with a single
        decision iteration with timesteps covering the entire model horizon.

        With a rule based approach, there might be many steps in the loop, each with a single
        decision iteration and single timestep, moving on once some threshold is satisfied.

        With a genetic algorithm, there might be a configurable number of steps in the loop,
        each with multiple decision iterations (one for each member of the algorithm's
        population) and timesteps covering the entire model horizon.

        Implicitly, if the bundle returned in an iteration contains multiple decision
        iterations, they can be performed in parallel. If each decision iteration contains
        multiple timesteps, they can also be parallelised, so long as there are no temporal
        dependencies.
        """
        assert len(self._decision_modules) == 1
        assert isinstance(self._decision_modules[0], PreSpecified)
        yield {0: self._timesteps}

    def get_state(self, timestep, iteration):
        """Return all interventions built in the given timestep for the given decision
        iteration.

        Deprecated - to be pushed down into DataHandle
        """
        self._decision_modules[0].get_state(timestep, iteration)


class DecisionModule(metaclass=ABCMeta):
    """Abstract class which provides the interface to decision mechanisms.

    These mechanisms including Pre-Specified Planning, a Rule-based Approach and
    Multi-objective Optimisation.

    This class provides two main public methods, ``__next__`` which is normally
    called implicitly as a call to the class as an iterator, and ``get_state()``
    which takes as arguments a smif.model.Model object, and ``timestep`` and
    ``decision_iteration`` integers. The first of these returns a dict of
    decision_iterations and timesteps over which a SosModel should be iterated.
    The latter provides a means to furnish the structure of contained Model
    objects through a list of historical and recent interventions.

    Arguments
    ---------
    timesteps : list
        A list of planning timesteps

    """
    def __init__(self, timesteps, register):
        self.timesteps = timesteps
        self.register = register

    def __next__(self):
        return self._get_next_decision_iteration()

    @abstractmethod
    def _get_next_decision_iteration(self):
        """Implement to return the next decision iteration

        Within a list of decision-iteration/timestep pairs, the assumption is
        that all decision iterations can be run in parallel
        (otherwise only one will be returned) and within a decision interation,
        all timesteps may be run in parallel as long as there are no
        inter-timestep state transitions required (e.g. reservoir level)

        Returns
        -------
        dict
            A dictionary keyed by decision iteration (int) whose values contain
            a list of timesteps
        """
        raise NotImplementedError

    @abstractmethod
    def get_state(self, timestep, iteration):
        """Return decisions for a given timestep and decision iteration
        """
        raise NotImplementedError

    def _get_previous_state(self, timestep, decision_iteration):
        """Gets state of the previous `timestep` for `decision_iteration`

        Arguments
        ---------
        timestep : int
        decision_iteration : int

        Returns
        -------
        numpy.ndarray
        """
        return self.get_state(timestep.PREVIOUS, decision_iteration)

    def _set_post_decision_state(self, timestep, decision_iteration, decision):
        """Sets the post-decision state

        Arguments
        ---------
        timestep : int
        decision_iteration : int
        decision : numpy.ndarray

        Notes
        -----
        `decision` should contain only the newly decided interventions
        """
        state = self._get_previous_state(timestep, decision_iteration)
        post_decision_state = state.bitwise_or(decision)
        self.register.set_state(timestep, decision_iteration, post_decision_state)

    @abstractmethod
    def _set_state(self, timestep, decision_iteration):
        """Implement to set the current state of a Model

        Arguments
        ---------
        timestep : int
        decision_iteration : int

        Notes
        -----
        1. Get state at previous timestep
        2. Compute decisions (interventions at current timestep)
        3. Accumulate to create current state
        4. Write this via the DataHandle to DataInterface

        This is a candidate for memoization

        """
        raise NotImplementedError


class PreSpecified(DecisionModule):
    """Pre-specified planning
    """

    def __init__(self, timesteps, register):
        super().__init__(timesteps, register)

    def _get_next_decision_iteration(self):
        return {1: [year for year in self.timesteps]}

    def _set_state(self, timestep, decision_iteration):
        """Pre-specified planning interventions are loaded during initialisation

        Pre-specified planning interventions are loaded during initialisation
        of a model run. This method just needs to copy the existing system state
        to the correct decision iteration reference.
        """
        pass

    def get_state(self, timestep, iteration=None):
        """Return a dict of intervention names built in timestep

        Returns
        -------
        dict

        Examples
        --------
        >>> dm = PreSpecified([2010, 2015], [{'name': 'intervention_a', 'build_year': 2010}])
        >>> dm.get_state(2010)
        {2010: ['intervention_a']}

        """
        state = {}

        for intervention in self.register:
            build_year = intervention['build_year']
            name = intervention['name']
            if self.buildable(build_year, timestep):
                if build_year in state:
                    state[build_year].append(name)
                else:
                    state[build_year] = [name]

        return state

    def buildable(self, build_year, timestep):
        """Interventions are deemed available if build_year is less than next timestep

        For example, if `a` is built in 2011 and timesteps are
        [2005, 2010, 2015, 2020] then buildable returns True for timesteps
        2010, 2015 and 2020 and False for 2005.
        """
        if timestep not in self.timesteps:
            raise ValueError("Timestep not in model timesteps")
        index = self.timesteps.index(timestep)
        if index == len(self.timesteps) - 1:
            next_year = timestep + 1
        else:
            next_year = self.timesteps[index + 1]

        if build_year < next_year:
            return True
        else:
            return False


class RuleBased(DecisionModule):
    """Rule-base decision modules
    """

    def __init__(self, timesteps):
        super().__init__(timesteps, register=None)
        self.satisfied = False
        self.current_timestep_index = 0
        self.current_iteration = 0

    def _get_next_decision_iteration(self):
            if self.satisfied and self.current_timestep_index == len(self.timesteps) - 1:
                return None
            elif self.satisfied and self.current_timestep_index <= len(self.timesteps):
                self.satisfied = False
                self.current_timestep_index += 1
                self.current_iteration += 1
                return {self.current_iteration: [self.timesteps[self.current_timestep_index]]}
            else:
                self.current_iteration += 1
                return {self.current_iteration: [self.timesteps[self.current_timestep_index]]}

    def get_state(self, timestep, iteration):
        return {}

    def _set_state(self, timestep, decision_iteration):
        raise NotImplementedError



"""Test command line interface
"""

import os
import subprocess
from tempfile import TemporaryDirectory
from unittest.mock import call, patch

import pkg_resources

import smif
from smif.cli import confirm, parse_arguments, setup_project_folder
from smif.controller.build import get_narratives
from smif.data_layer import DatafileInterface
from smif.parameters import Narrative


def get_args(args):
    """Get args object from list of strings
    """
    parser = parse_arguments()
    return parser.parse_args(args)


def test_parse_arguments():
    """Setup a project folder argument parsing
    """
    with TemporaryDirectory() as project_folder:
        args = get_args(['setup', '-d', project_folder])

        expected = project_folder
        actual = args.directory
        assert actual == expected

        # Ensure that the `setup_project_folder` function is called when `setup`
        # command is passed to the cli
        assert args.func.__name__ == 'setup_project_folder'


def test_fixture_single_run():
    """Test running the (default) binary-filesystem-based single_run fixture
    """
    config_dir = pkg_resources.resource_filename('smif', 'sample_project')
    output = subprocess.run(["smif", "-v", "run", "-d", config_dir,
                             "20170918_energy_water_short"],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    assert "Running 20170918_energy_water_short" in str(output.stderr)
    assert "Model run '20170918_energy_water_short' complete" in str(output.stdout)


def test_fixture_single_run_csv():
    """Test running the csv-filesystem-based single_run fixture
    """
    config_dir = pkg_resources.resource_filename('smif', 'sample_project')
    output = subprocess.run(["smif", "-v", "run", "-i", "local_csv", "-d", config_dir,
                             "20170918_energy_water_short"],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    assert "Running 20170918_energy_water_short" in str(output.stderr)
    assert "Model run '20170918_energy_water_short' complete" in str(output.stdout)


def test_fixture_single_run_warm():
    """Test running the (default) single_run fixture with warm setting enabled
    """
    config_dir = pkg_resources.resource_filename('smif', 'sample_project')
    output = subprocess.run(["smif", "-v", "run", "-w", "-d", config_dir,
                             "20170918_energy_water_short"],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    assert "Running 20170918_energy_water_short" in str(output.stderr)
    assert "Model run '20170918_energy_water_short' complete" in str(output.stdout)


def test_fixture_batch_run():
    """Test running the multiple modelruns using the batch_run option
    """
    config_dir = pkg_resources.resource_filename('smif', 'sample_project')
    output = subprocess.run(["smif", "-v", "run", "-b", "-d", config_dir,
                             os.path.join(config_dir, "batchfile")],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    assert "Running 20170918_energy_water" in str(output.stderr)
    assert "Model run '20170918_energy_water' complete" in str(output.stdout)
    assert "Running 20170918_energy_water_short" in str(output.stderr)
    assert "Model run '20170918_energy_water_short' complete" in str(output.stdout)


def test_fixture_list_runs():
    """Test running the filesystem-based single_run fixture
    """
    config_dir = pkg_resources.resource_filename('smif', 'sample_project')
    output = subprocess.run(["smif", "list", "-d", config_dir], stdout=subprocess.PIPE)
    assert "20170918_energy_water" in str(output.stdout)
    assert "20170918_energy_water_short" in str(output.stdout)


def test_setup_project_folder():
    """Test contents of the setup project folder
    """
    with TemporaryDirectory() as project_folder:
        args = get_args(['setup', '-d', project_folder])
        setup_project_folder(args)

        assert os.path.exists(project_folder)

        folder_list = ['config', 'data', 'models', 'planning']
        for folder in folder_list:
            folder_path = os.path.join(project_folder, folder)

            assert os.path.exists(folder_path)


@patch('builtins.input', return_value='y')
def test_confirm_yes(input):
    assert confirm()


@patch('builtins.input', return_value='n')
def test_confirm_no(input):
    assert not confirm()


@patch('builtins.input', return_value='')
def test_confirm_default_response(input):
    assert not confirm()


@patch('builtins.input', return_value='n')
@patch('builtins.print')
def test_confirm_default_message(mock_print, input):
    confirm()
    input.assert_has_calls([call('Confirm [n]|y: ')])


@patch('builtins.input', return_value='n')
@patch('builtins.print')
def test_confirm_custom_message(mock_print, input):
    confirm('Create directory?', True)
    input.assert_has_calls([call('Create directory? [y]|n: ')])


@patch('builtins.input', side_effect=['invalid', 'y'])
@patch('builtins.print')
def test_confirm_repeat_message(mock_print, input):
    confirm()
    input.assert_has_calls([call('Confirm [n]|y: '), call('Confirm [n]|y: ')])
    mock_print.assert_called_with('please enter y or n.')


def test_version_display():
    """Expect version number from `smif -V`
    """
    output = subprocess.run(['smif', '-V'], stdout=subprocess.PIPE)
    assert smif.__version__ in str(output.stdout)


def test_verbose_debug():
    """Expect debug message from `smif -vv`
    """
    output = subprocess.run(['smif', '-vv'], stderr=subprocess.PIPE)
    assert 'DEBUG' in str(output.stderr)


def test_verbose_debug_alt():
    """Expect debug message from `smif --verbose --verbose`
    """
    output = subprocess.run(['smif', '--verbose', '--verbose'], stderr=subprocess.PIPE)
    assert 'DEBUG' in str(output.stderr)


def test_verbose_info(setup_folder_structure):
    """Expect info message from `smif -v validate <config_file>`
    """
    config_file = os.path.join(str(setup_folder_structure))
    output = subprocess.run(['smif', '-v', 'run', config_file], stderr=subprocess.PIPE)
    assert 'INFO' in str(output.stderr)


class TestRunSosModelRunComponents():

    def test_get_narratives(self):
        """should load a list of narratives with parameter value data
        """
        config_dir = pkg_resources.resource_filename('smif', 'sample_project')
        handler = DatafileInterface(config_dir, 'local_csv')
        narratives = {'technology': ['High Tech Demand Side Management']}
        actual = get_narratives(handler, narratives)

        data = {
            'energy_demand': {
                'smart_meter_savings': 8
            },
            'water_supply': {
                'clever_water_meter_savings': 8,
                'per_capita_water_demand': 1.2
            }
        }
        name = 'High Tech Demand Side Management'
        description = 'High penetration of SMART technology on the demand side'
        narrative_set = 'technology'

        narrative_object = Narrative(name, description, narrative_set)
        narrative_object.data = data

        assert actual == [narrative_object]



"""Tests the definition and solution of the planning problem optimisation
"""
from skopt import dummy_minimize
from smif.optimisation import OptimizerTemplate


class Optimizer(OptimizerTemplate):
    """A simple algorithm which minimizes the optimization function
    """

    def run(self):

        dimensions = [(0, 1)] * self._action_space
        results = dummy_minimize(self.optimization_function,
                                 dimensions,
                                 random_state=1,
                                 n_calls=1000,
                                 verbose=True
                                 )
        self.results['optimal_decision'] = results.x
        self.results['objective_value'] = results.fun


def dummy_function(binary_vector):
    """A dummy function, representing a system-of-systems model

    The function accepts a list of 0-1 intervention decisions and returns
    a scalar

    Arguments
    ---------
    binary_vector : list
        A list of 0-1 values

    """

    def binary_to_integer(binary_vector):
        """Convert list of binary integers to integer scalar
        """
        binary_string = "".join(str(x) for x in binary_vector)
        return float(int('0b{}'.format(binary_string), 2))

    scalar = binary_to_integer(binary_vector)

    return scalar ** 2


class TestOptimizationImplementation:

    def test_optimise_function(self):
        """Demonstrates the steps to run the optimization algorithm
        """

        optimiser = Optimizer()
        number_dimensions = 8  # Action space
        optimiser.initialize(number_dimensions)
        optimiser.optimization_function = dummy_function
        optimiser.run()
        assert optimiser.results['optimal_decision'] == [0, 0, 0, 0, 0, 0, 0, 0]
        assert optimiser.results['objective_value'] == 0


class TestOptimizationTemplate:

    def test_implement_template(self):

        class MyOpt(OptimizerTemplate):

            def run(self):
                self.results['optimal_decision'] = [1] * self._action_space
                self.results['objective_value'] = 0

        myopt = MyOpt()
        myopt.initialize(1)
        assert myopt._action_space == 1

        myopt.optimization_function = lambda x: x

        myopt.run()

        assert myopt.results == {'optimal_decision': [1],
                                 'objective_value': 0}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""smif
"""
from __future__ import division, print_function, absolute_import

import pkg_resources

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


try:
    __version__ = pkg_resources.get_distribution(__name__).version
except:
    __version__ = 'unknown'


class StateData(object):
    """A piece of state data

    Parameters
    ----------
    target
        The id or name of the object described by this state
    data
        The state attribute/data to apply - could typically be a dict of
        attributes
    """
    def __init__(self, target, data):
        self.target = target
        self.data = data

    def __repr__(self):
        return "StateData({}, {})".format(
            repr(self.target),
            repr(self.data)
        )



"""Test ModelData
"""
from unittest.mock import MagicMock, Mock

import numpy as np
from pytest import fixture, raises
from smif.data_layer import DataHandle
from smif.data_layer.data_handle import TimestepResolutionError
from smif.metadata import Metadata, MetadataSet
from smif.model.dependency import Dependency
from smif.parameters import ParameterList


@fixture(scope='function')
def mock_store():
    store = Mock()
    store.read_results = MagicMock(return_value=np.array([[1.0]]))
    return store


@fixture(scope='function')
def mock_model():
    model = Mock()
    model.name = 'test_model'
    model.parameters = ParameterList()
    model.parameters.add_parameter(
        {
            'name': 'smart_meter_savings',
            'description': 'The savings from smart meters',
            'absolute_range': (0, 100),
            'suggested_range': (3, 10),
            'default_value': 3,
            'units': '%',
            'parent': None
        }
    )
    regions = Mock()
    regions.name = 'half_squares'
    regions.__len__ = lambda self: 1
    intervals = Mock()
    intervals.name = 'remap_months'
    intervals.__len__ = lambda self: 1
    model.inputs = MetadataSet([
        Metadata('test', regions, intervals, 'm')
    ])
    model.outputs = MetadataSet([
        Metadata('test', regions, intervals, 'm')
    ])
    source_model = Mock()
    source_model.name = 'test_source'
    model.deps = {
        'test': Dependency(
            source_model,
            Metadata('test_output', regions, intervals, 'm'),
            Metadata('test', regions, intervals, 'm')
        )
    }
    return model


@fixture(scope='function')
def mock_model_with_conversion():
    model = Mock()
    model.name = 'test_model'
    model.parameters = ParameterList()
    model.parameters.add_parameter(
        {
            'name': 'smart_meter_savings',
            'description': 'The savings from smart meters',
            'absolute_range': (0, 100),
            'suggested_range': (3, 10),
            'default_value': 3,
            'units': '%',
            'parent': None
        }
    )
    regions = Mock()
    regions.name = 'half_squares'
    intervals = Mock()
    intervals.name = 'remap_months'
    model.inputs = MetadataSet([
        Metadata('test', regions, intervals, 'liters')
    ])
    source_model = Mock()
    source_model.name = 'test_source'
    model.deps = {
        'test': Dependency(
            source_model,
            Metadata('test_output', regions, intervals, 'milliliters'),
            Metadata('test', regions, intervals, 'liters')
        )
    }
    return model


class TestDataHandle():
    def test_create(self):
        """should be created with a DataInterface
        """
        DataHandle(Mock(), 1, 2015, [2015, 2020], Mock())

    def test_get_data(self, mock_store, mock_model):
        """should allow read access to input data
        """
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)
        expected = np.array([[1.0]])
        actual = data_handle.get_data("test")
        assert actual == expected

        mock_store.read_results.assert_called_with(
            1,
            'test_source',  # read from source model
            'test_output',  # using source model output name
            'half_squares',
            'remap_months',
            2015,
            None,
            None
        )

    def test_get_data_with_conversion(self, mock_store, mock_model_with_conversion):
        """should convert liters to milliliters (1 -> 0.001)
        """
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model_with_conversion)
        expected = np.array([[0.001]])
        actual = data_handle.get_data("test")
        assert actual == expected

        mock_store.read_results.assert_called_with(
            1,
            'test_source',  # read from source model
            'test_output',  # using source model output name
            'half_squares',
            'remap_months',
            2015,
            None,
            None
        )

    def test_get_base_timestep_data(self, mock_store, mock_model):
        """should allow read access to input data from base timestep
        """
        data_handle = DataHandle(mock_store, 1, 2025, [2015, 2020, 2025], mock_model)
        expected = np.array([[1.0]])
        actual = data_handle.get_base_timestep_data("test")
        assert actual == expected

        mock_store.read_results.assert_called_with(
            1,
            'test_source',  # read from source model
            'test_output',  # using source model output name
            'half_squares',
            'remap_months',
            2015,  # base timetep
            None,
            None
        )

    def test_get_previous_timestep_data(self, mock_store, mock_model):
        """should allow read access to input data from previous timestep
        """
        data_handle = DataHandle(mock_store, 1, 2025, [2015, 2020, 2025], mock_model)
        expected = np.array([[1.0]])
        actual = data_handle.get_previous_timestep_data("test")
        assert actual == expected

        mock_store.read_results.assert_called_with(
            1,
            'test_source',  # read from source model
            'test_output',  # using source model output name
            'half_squares',
            'remap_months',
            2020,  # previous timetep
            None,
            None
        )

    def test_get_data_with_square_brackets(self, mock_store, mock_model):
        """should allow dict-like read access to input data
        """
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)
        expected = np.array([[1.0]])
        actual = data_handle["test"]
        assert actual == expected

        mock_store.read_results.assert_called_with(
            1,
            'test_source',  # read from source model
            'test_output',  # using source model output name
            'half_squares',
            'remap_months',
            2015,
            None,
            None
        )

    def test_set_data(self, mock_store, mock_model):
        """should allow write access to output data
        """
        expected = np.array([[1.0]])
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)

        data_handle.set_results("test", expected)
        assert data_handle["test"] == expected

        mock_store.write_results.assert_called_with(
            1,
            'test_model',
            'test',
            expected,
            'half_squares',
            'remap_months',
            2015,
            None,
            None
        )

    def test_set_data_wrong_shape(self, mock_store, mock_model):
        """should allow write access to output data
        """
        expect_error = np.array([[1.0, 1.0]])  # regions is 1, intervals is 1 not 2
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)

        with raises(ValueError) as ex:
            data_handle.set_results("test", expect_error)

        assert "Tried to set results with shape (1, 2), " + \
            "expected (1, 1) for test_model:test" in str(ex)

    def test_set_data_with_square_brackets(self, mock_store, mock_model):
        """should allow dict-like write access to output data
        """
        expected = np.array([[1.0]])
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)

        data_handle["test"] = expected
        assert data_handle["test"] == expected

        mock_store.write_results.assert_called_with(
            1,
            'test_model',
            'test',
            expected,
            'half_squares',
            'remap_months',
            2015,
            None,
            None
        )

    def test_get_regions(self, mock_store, mock_model):
        """should allow read access to input data
        """
        mock_store.read_region_names = Mock(return_value=['a', 'b'])
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)
        expected = ['a', 'b']
        actual = data_handle.get_region_names("half_squares")
        assert actual == expected

        mock_store.read_region_names.assert_called_with(
            'half_squares')

    def test_get_intervals(self, mock_store, mock_model):
        """should allow read access to input data
        """
        mock_store.read_interval_names = Mock(return_value=['a', 'b'])
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)
        expected = ['a', 'b']
        actual = data_handle.get_interval_names("remap_months")
        assert actual == expected

        mock_store.read_interval_names.assert_called_with(
            'remap_months')


class TestDataHandleTimesteps():
    """Test timestep helper properties
    """
    def test_current_timestep(self):
        """should return current timestep
        """
        data_handle = DataHandle(Mock(), 1, 2015, [2015, 2020], Mock())
        assert data_handle.current_timestep == 2015

    def test_base_timestep(self):
        """should return first timestep in list
        """
        data_handle = DataHandle(Mock(), 1, 2015, [2015, 2020], Mock())
        assert data_handle.base_timestep == 2015

    def test_previous_timestep(self):
        """should return previous timestep from list
        """
        data_handle = DataHandle(Mock(), 1, 2020, [2015, 2020], Mock())
        assert data_handle.previous_timestep == 2015

    def test_previous_timestep_error(self):
        """should raise error if there's no previous timestep in the list
        """
        data_handle = DataHandle(Mock(), 1, 2015, [2015, 2020], Mock())
        with raises(TimestepResolutionError) as ex:
            data_handle.previous_timestep
        assert 'no previous timestep' in str(ex)



from unittest.mock import Mock

import numpy as np
from smif.convert.unit import get_register


def test_register_user_unit(setup_folder_structure):
    """Test that we can load in user defined units
    """
    test_folder = setup_folder_structure
    units_file = test_folder.join('data', 'user_units.txt')

    register = get_register()
    register.register(str(units_file))

    assert register.parse_unit('people') == 'people'


def test_parse_unit_valid():
    """Parse a valid unit
    """
    register = get_register()
    meter = register.parse_unit('m')
    assert str(meter) == 'meter'


def test_parse_unit_invalid():
    """Warn if unit not recognised
    """
    unit = 'unrecognisable'
    register = get_register()
    register.LOGGER.warning = Mock()
    register.parse_unit(unit)
    msg = "Unrecognised unit: %s"
    register.LOGGER.warning.assert_called_with(msg, unit)


def test_convert_unit():

    data = np.array([[1, 2], [3, 4]], dtype=float)

    register = get_register()
    actual = register.convert(data, 'liter', 'milliliter')

    expected = np.array([[1000, 2000], [3000, 4000]], dtype=float)

    np.allclose(actual, expected)



from pytest import fixture, raises
from smif.intervention import Intervention, InterventionRegister


@fixture(scope='function')
def get_intervention():
    name = 'water_treatment_plant'
    data = {
        'capacity': {
            'units': 'ML/day',
            'value': 5
        },
        'location': 'oxford'
    }
    return Intervention(
        name=name,
        data=data,
        sector='water_supply'
    )


@fixture(scope='function')
def build_intervention_ws():
    data = {
        'sector': 'water_supply',
        'name': 'oxford treatment plant',
        'capacity': {
            'units': 'ML/day',
            'value': 450
            },
        'capital_cost': {
            'units': 'M£',
            'value': 500
        },
        'location': "POINT(51.1 -1.7)"
        }

    return data


@fixture(scope='function')
def build_intervention_es():
    data = {
        'sector': 'energy_supply',
        'name': 'London Array',
        'capacity': {
            'units': 'GW',
            'value': 2
            },
        'capital_cost': {
            'units': 'M£',
            'value': 2500
        },
        'location': "POINT(50.1 -1.4)"
        }

    return data


@fixture(scope='function')
def build_register_two(build_intervention_es, build_intervention_ws):
    water = Intervention(data=build_intervention_ws)
    energy = Intervention(data=build_intervention_es)

    register = InterventionRegister()
    register.register(energy)
    register.register(water)
    return register


class TestIntervention:

    def test_intervention_init_sector(self, build_intervention_ws):
        actual = Intervention(data=build_intervention_ws)
        assert actual.sector == 'water_supply'

    def test_intervention_init_build_date(self, build_intervention_ws):
        actual = Intervention(data=build_intervention_ws)
        with raises(AttributeError):
            assert actual.build_date

    def test_intervention_init_location(self, build_intervention_ws):
        actual = Intervention(data=build_intervention_ws)
        assert actual.location == "POINT(51.1 -1.7)"


class TestInterventionRegister:

    def test_register_intervention(self, get_intervention):
        water_treatment_plant = get_intervention
        register = InterventionRegister()
        register.register(water_treatment_plant)

        assert len(register) == 1
        assert sorted(register._attribute_keys) == [
            "capacity",
            "location",
            "name",
            "sector"
        ]

        attr_idx = register.attribute_index("name")
        possible = register._attribute_possible_values[attr_idx]
        assert possible == [None, "water_treatment_plant"]

        attr_idx = register.attribute_index("capacity")
        possible = register._attribute_possible_values[attr_idx]
        assert possible == [None, {'units': 'ML/day', 'value': 5}]

    def test_retrieve_intervention(self, get_intervention):
        water_treatment_plant = get_intervention
        register = InterventionRegister()
        register.register(water_treatment_plant)

        # pick an asset from the list - this is what the optimiser will do
        numeric_asset = register._numeric_keys[0]

        asset = register.numeric_to_intervention(numeric_asset)
        assert asset.name == "water_treatment_plant"

        assert asset.data == {
            'name': 'water_treatment_plant',
            'sector': 'water_supply',
            'capacity': {
                'units': 'ML/day',
                'value': 5
            },
            'location': 'oxford'
        }

    def test_retrieve_intervention_by_name(self, get_intervention):
        water_treatment_plant = get_intervention
        register = InterventionRegister()
        register.register(water_treatment_plant)

        # pick an asset from the list - this is what the optimiser will do
        asset = register.get_intervention('water_treatment_plant')
        assert asset.name == "water_treatment_plant"

        assert asset.data == {
            'name': 'water_treatment_plant',
            'sector': 'water_supply',
            'capacity': {
                'units': 'ML/day',
                'value': 5
            },
            'location': 'oxford'
        }

    def test_error_when_retrieve_by_name(self, get_intervention):
        water_treatment_plant = get_intervention
        register = InterventionRegister()
        register.register(water_treatment_plant)
        with raises(ValueError) as excinfo:
            register.get_intervention('not_here')
        expected = "Intervention 'not_here' not found in register"
        assert str(excinfo.value) == expected

    def test_iterate_over_interventions(self, get_intervention):
        """Test __iter___ method of AssetRegister class

        """
        asset_one = get_intervention
        register = InterventionRegister()
        register.register(asset_one)

        for asset in register:
            assert asset.sha1sum() == asset_one.sha1sum()

    def test_add_duplicate_intervention(self, get_intervention):
        """Tests that only unique interventions are retained

        """
        asset_one = get_intervention
        asset_two = get_intervention
        register = InterventionRegister()
        register.register(asset_one)
        register.register(asset_two)

        assert len(register) == 1

    def test_register_len_one(self, build_intervention_ws):
        water = Intervention(data=build_intervention_ws)
        register = InterventionRegister()
        register.register(water)
        assert len(register) == 1

    def test_register_len_two(self, build_register_two):
        register = build_register_two
        assert len(register) == 2




import numpy as np
from pytest import fixture
from smif.data_layer import DataHandle, MemoryInterface
from smif.model.sos_model import ModelSet, SectorModel


@fixture(scope='function')
def get_empty_sector_model():
    """Return model with static outputs
    """
    class EmptySectorModel(SectorModel):
        """Static water supply model
        """
        def initialise(self, initial_conditions):
            pass

        def simulate(self, data):
            data['cost'] = np.array([[2]])
            data['water'] = np.array([[0.5]])
            return data

        def extract_obj(self, results):
            return 0

    return EmptySectorModel


@fixture(scope='function')
def get_sector_model_object(get_empty_sector_model):
    """Return model configured with two outputs, no inputs
    """
    sector_model = get_empty_sector_model('water_supply')

    regions = sector_model.regions
    intervals = sector_model.intervals

    sector_model.add_output('cost',
                            regions.get_entry('LSOA'),
                            intervals.get_entry('annual'),
                            'million GBP')

    sector_model.add_output('water',
                            regions.get_entry('LSOA'),
                            intervals.get_entry('annual'),
                            'Ml')

    return sector_model


def get_data_handle(model):
    """Return a data handle for the model
    """
    store = MemoryInterface()
    store.write_sos_model_run({
        'name': 'test',
        'narratives': {}
    })
    return DataHandle(
        store,
        'test',  # modelrun_name
        2010,  # current_timestep
        [2010, 2011],  # timesteps
        model
    )


class TestModelSet:
    """Test public interface to ModelSet, with side effects in DataHandle
    """
    def test_guess_outputs_zero(self, get_sector_model_object):
        """If no previous timestep has results, guess outputs as zero
        """
        ws_model = get_sector_model_object
        model_set = ModelSet({ws_model.name: ws_model})
        data_handle = get_data_handle(model_set)
        model_set.simulate(data_handle)
        expected = {
            "cost": np.zeros((1, 1)),
            "water": np.zeros((1, 1))
        }
        actual = {
            "cost": data_handle.get_results('cost', ws_model.name, 0),
            "water": data_handle.get_results('water', ws_model.name, 0)
        }
        assert actual == expected

    def test_guess_outputs_last_year(self, get_sector_model_object):
        """If a previous timestep has results, guess outputs as identical
        """
        ws_model = get_sector_model_object
        model_set = ModelSet({ws_model.name: ws_model})

        expected = {
            "cost": np.array([[3.14]]),
            "water": np.array([[2.71]])
        }

        # set up data as though from previous timestep simulation
        data_handle = get_data_handle(model_set)
        data_handle._store.write_results(
            'test', 'water_supply', 'cost', expected['cost'], 'LSOA', 'annual', 2010, 0)
        data_handle._store.write_results(
            'test', 'water_supply', 'water', expected['water'], 'LSOA', 'annual', 2010, 0)

        data_handle._current_timestep = 2011
        model_set.simulate(data_handle)
        actual = {
            "cost": data_handle.get_results('cost', ws_model.name, 0),
            "water": data_handle.get_results('water', ws_model.name, 0)
        }
        assert actual == expected

    def test_converged_first_iteration(self, get_sector_model_object):
        """Should not report convergence after a single iteration
        """
        ws_model = get_sector_model_object
        model_set = ModelSet({ws_model.name: ws_model})
        data_handle = get_data_handle(model_set)
        model_set.simulate(data_handle)
        assert model_set.max_iteration > 0

    def test_converged_two_identical(self, get_sector_model_object):
        """Should report converged if the last two output sets are identical
        """
        ws_model = get_sector_model_object
        model_set = ModelSet({ws_model.name: ws_model})

        data_handle = get_data_handle(model_set)
        model_set.simulate(data_handle)
        # ws_model will always return identical results, so we expect:
        # 0: guessed zeroes
        # 1: returned values
        # 2: returned identical values
        assert model_set.max_iteration == 2



# -*- coding: utf-8 -*-

from copy import copy
from unittest.mock import Mock

import numpy as np
import pytest
from pytest import fixture, raises
from smif.metadata import MetadataSet
from smif.model.dependency import Dependency
from smif.model.scenario_model import ScenarioModel
from smif.model.sector_model import SectorModel
from smif.model.sos_model import SosModel, SosModelBuilder

from ..fixtures.water_supply import WaterSupplySectorModel


@fixture(scope='function')
def get_scenario_model_object():

    scenario_model = ScenarioModel('test_scenario_model')
    scenario_model.add_output('raininess',
                              scenario_model.regions.get_entry('LSOA'),
                              scenario_model.intervals.get_entry('annual'),
                              'ml')
    scenario_model.scenario_set = 'raininess'
    return scenario_model


@fixture(scope='function')
def get_sector_model_object(get_empty_sector_model):

    sector_model = get_empty_sector_model('water_supply')

    regions = sector_model.regions
    intervals = sector_model.intervals

    sector_model.add_input('raininess',
                           regions.get_entry('LSOA'),
                           intervals.get_entry('annual'),
                           'ml')

    sector_model.add_output('cost',
                            regions.get_entry('LSOA'),
                            intervals.get_entry('annual'),
                            'million GBP')

    sector_model.add_output('water',
                            regions.get_entry('LSOA'),
                            intervals.get_entry('annual'),
                            'Ml')

    return sector_model


@fixture(scope='function')
def get_sos_model_object(get_sector_model_object,
                         get_scenario_model_object):

    sos_model = SosModel('test_sos_model')
    sector_model = get_sector_model_object
    scenario_model = get_scenario_model_object
    sos_model.add_model(sector_model)
    sos_model.add_model(scenario_model)
    sector_model.add_dependency(scenario_model, 'raininess', 'raininess')

    return sos_model


@fixture(scope='function')
def get_sos_model_with_summed_dependency(oxford_region):
    scenario_model = get_scenario_model_object

    builder = SosModelBuilder()
    builder.load_scenario_models([scenario_model])

    sos_model = builder.finish()

    region_register = sos_model.regions
    interval_register = sos_model.intervals

    raininess_model = sos_model.models['raininess']

    ws = WaterSupplySectorModel('water_supply')
    ws.add_input(
        'raininess',
        region_register.get_entry('LSOA'),
        interval_register.get_entry('annual'),
        'ml')
    ws.add_output(
        'water',
        region_register.get_entry('LSOA'),
        interval_register.get_entry('annual'),
        'Ml')
    ws.add_dependency(raininess_model, 'raininess', 'raininess')
    sos_model.add_model(ws)

    ws2 = WaterSupplySectorModel('water_supply_2')
    ws2.add_input(
        'raininess',
        region_register.get_entry('LSOA'),
        interval_register.get_entry('annual'),
        'ml')
    ws2.add_output(
        'water',
        region_register.get_entry('LSOA'),
        interval_register.get_entry('annual'),
        'Ml')
    ws2.add_dependency(raininess_model, 'raininess', 'raininess')
    sos_model.add_model(ws2)

    ws3 = WaterSupplySectorModel('water_supply_3')
    ws3.add_input(
        'water',
        region_register.get_entry('LSOA'),
        interval_register.get_entry('annual'),
        'Ml')
    # TODO implement summed dependency
    # ws3.add_dependency(ws, 'water', 'water')
    ws3.add_dependency(ws2, 'water', 'water')
    sos_model.add_model(ws3)

    return sos_model


@fixture(scope='function')
def get_empty_sector_model():

    class EmptySectorModel(SectorModel):

        def initialise(self, initial_conditions):
            pass

        def simulate(self, data):
            return data

        def extract_obj(self, results):
            return 0

    return EmptySectorModel


class TestSosModelProperties():

    def test_model_inputs_property(self, get_sos_model_object):
        sos_model = get_sos_model_object

        expected = {'raininess': 'water_supply'}

        for key, value in expected.items():
            inputs = sos_model.models[value].inputs
            assert isinstance(inputs, MetadataSet)
            assert key in inputs.names

    def test_model_outputs_property(self, get_sos_model_object):
        sos_model = get_sos_model_object

        expected = {'cost': 'water_supply'}

        for key, value in expected.items():
            outputs = sos_model.models[value].outputs
            assert isinstance(outputs, MetadataSet)
            assert key in outputs.names


class TestSosModel():

    def test_serialise_configuration(self, get_sos_model_object):
        """Tests that as_dict function correctly returns configuration
        as a dictionary
        """
        sos_model = get_sos_model_object
        actual = sos_model.as_dict()
        expected = {
            'name': 'test_sos_model',
            'description': '',
            'scenario_sets': ['raininess'],
            'sector_models': ['water_supply'],
            'dependencies': [{
                'source_model': 'test_scenario_model',
                'source_model_output': 'raininess',
                'sink_model': 'water_supply',
                'sink_model_input': 'raininess'
            }],
            'max_iterations': 25,
            'convergence_absolute_tolerance': 1e-8,
            'convergence_relative_tolerance': 1e-5
        }
        assert actual == expected

    def test_run_with_global_parameters(self, get_sos_model_object):
        sos_model = get_sos_model_object
        sos_model.add_parameter({
            'name': 'sos_model_param',
            'description': 'A global parameter passed to all contained models',
            'absolute_range': (0, 100),
            'suggested_range': (3, 10),
            'default_value': 3,
            'units': '%'
        })
        assert 'sos_model_param' in sos_model.parameters

    def test_run_with_sector_parameters(self, get_sos_model_object):
        sos_model = get_sos_model_object
        sector_model = sos_model.models['water_supply']
        sector_model.add_parameter({
            'name': 'sector_model_param',
            'description': 'A model parameter passed to a specific model',
            'absolute_range': (0, 100),
            'suggested_range': (3, 10),
            'default_value': 3,
            'units': '%'
        })
        assert 'sector_model_param' in sector_model.parameters

    def test_add_parameters(self, get_empty_sector_model):
        sos_model = SosModel('global')
        sos_model_param = {
            'name': 'sos_model_param',
            'description': 'A global parameter passed to all contained models',
            'absolute_range': (0, 100),
            'suggested_range': (3, 10),
            'default_value': 3,
            'units': '%'
        }
        sos_model.add_parameter(sos_model_param)
        expected = sos_model_param

        assert sos_model.parameters['sos_model_param'].as_dict() == expected
        assert sos_model.parameters.names == ['sos_model_param']

        sector_model = get_empty_sector_model('source_model')
        sector_model.add_parameter({
            'name': 'sector_model_param',
            'description': 'Required for the sectormodel',
            'absolute_range': (0, 100),
            'suggested_range': (3, 10),
            'default_value': 3,
            'units': '%'
        })
        sos_model.add_model(sector_model)

        # SosModel contains only its own parameters
        assert 'sos_model_param' in sos_model.parameters.names

        # SectorModel has its own ParameterList, gettable by param name
        assert 'sector_model_param' in sector_model.parameters.names

    def test_add_dependency(self, get_empty_sector_model):

        regions = Mock()
        regions.name = 'test_regions'
        intervals = Mock()
        intervals.name = 'test_intervals'
        units = 'test_units'

        sink_model = get_empty_sector_model('sink_model')
        sink_model.add_input('input_name', regions, intervals, units)

        source_model = get_empty_sector_model('source_model')
        source_model.add_output('output_name', regions, intervals, units)

        sink_model.add_dependency(source_model, 'output_name', 'input_name')

        sos_model = SosModel('test')
        sos_model.add_model(source_model)
        sos_model.add_model(sink_model)

    def test_timestep_before(self):
        sos_model = SosModel('test')
        sos_model.timesteps = [2010, 2011, 2012]
        assert sos_model.timestep_before(2010) is None
        assert sos_model.timestep_before(2011) == 2010
        assert sos_model.timestep_before(2012) == 2011
        assert sos_model.timestep_before(2013) is None

    def test_timestep_after(self):
        sos_model = SosModel('test')
        sos_model.timesteps = [2010, 2011, 2012]
        assert sos_model.timestep_after(2010) == 2011
        assert sos_model.timestep_after(2011) == 2012
        assert sos_model.timestep_after(2012) is None
        assert sos_model.timestep_after(2013) is None

    def test_run_sequential(self, get_sos_model_object):
        sos_model = get_sos_model_object
        data_handle = Mock()
        data_handle.get_state = Mock(return_value={})
        data_handle.timesteps = [2010, 2011, 2012]
        data_handle._current_timestep = 2010
        sos_model.simulate(data_handle)
        data_handle._current_timestep = 2011
        sos_model.simulate(data_handle)
        data_handle._current_timestep = 2012
        sos_model.simulate(data_handle)

    @pytest.mark.xfail(reason="Summed dependencies not yet implemented")
    def test_dependency_aggregation(self, get_sos_model_with_summed_dependency):
        sos_model = get_sos_model_with_summed_dependency

        data = {
            2010: {
                'decisions': [],
                'raininess': np.array([[1, 1]]),
                'water': np.array([2, 2])
            }
        }

        sos_model.simulate(2010, data)


@fixture(scope='function')
def get_sos_model_config(get_scenario_model_object):

    scenario_model = get_scenario_model_object
    config_data = {
        'name': 'energy_sos_model',
        'description': 'description of a sos model',
        'scenario_sets': [scenario_model],
        'sector_models': [],
        'dependencies': []
    }

    return config_data


@fixture
def get_sos_model_config_with_dep(get_sos_model_config,
                                  get_sector_model_object):

    sector_model = get_sector_model_object
    dependency_config = [{'source_model': 'test_scenario_model',
                          'source_model_output': 'raininess',
                          'sink_model': 'water_supply',
                          'sink_model_input': 'raininess'}]

    config_data = get_sos_model_config
    config_data['sector_models'].append(sector_model)
    config_data['dependencies'] = dependency_config

    return config_data


@fixture
def get_sos_model_config_with_summed_dependency(get_sos_model_config, get_sector_model_object):

    config = get_sos_model_config

    water_model_one = get_sector_model_object
    config['sector_models'].append(water_model_one)

    water_model_two = copy(config['sector_models'][0])
    water_model_two.name = 'water_supply_two'

    regions = water_model_two.regions
    intervals = water_model_two.intervals

    water_model_three = copy(config['sector_models'][0])
    water_model_three.name = 'water_supply_three'

    water_model_three.add_input('water',
                                regions.get_entry('LSOA'),
                                intervals.get_entry('annual'),
                                'Ml')
    config['dependencies'] = [{'source_model': 'water_supply',
                               'source_model_output': 'water',
                               'sink_model': 'water_supply_three',
                               'sink_model_input': 'water'},
                              {'source_model': 'water_supply_two',
                               'source_model_output': 'water',
                               'sink_model': 'water_supply_three',
                               'sink_model_input': 'water'},
                              {'source_model': 'test_scenario_model',
                               'source_model_output': 'raininess',
                               'sink_model': 'water_supply',
                               'sink_model_input': 'raininess'},
                              {'source_model': 'test_scenario_model',
                               'source_model_output': 'raininess',
                               'sink_model': 'water_supply_two',
                               'sink_model_input': 'raininess'}]
    config['sector_models'].append(water_model_two)
    config['sector_models'].append(water_model_three)
    return config


class TestSosModelBuilderComponents():

    def test_construct(self, get_sos_model_config):
        """Test constructing from single dict config
        """
        config = get_sos_model_config
        builder = SosModelBuilder()
        builder.construct(config)
        sos_model = builder.finish()

        assert isinstance(sos_model, SosModel)
        assert list(sos_model.scenario_models.keys()) == ['test_scenario_model']
        assert isinstance(sos_model.models['test_scenario_model'], ScenarioModel)

    def test_set_max_iterations(self, get_sos_model_config):
        """Test constructing from single dict config
        """
        config = get_sos_model_config
        config['max_iterations'] = 125
        builder = SosModelBuilder()
        builder.construct(config)
        sos_model = builder.finish()
        assert sos_model.max_iterations == 125

    def test_set_convergence_absolute_tolerance(self, get_sos_model_config):
        """Test constructing from single dict config
        """
        config = get_sos_model_config
        config['convergence_absolute_tolerance'] = 0.0001
        builder = SosModelBuilder()
        builder.construct(config)
        sos_model = builder.finish()
        assert sos_model.convergence_absolute_tolerance == 0.0001

    def test_set_convergence_relative_tolerance(self, get_sos_model_config):
        """Test constructing from single dict config
        """
        config = get_sos_model_config
        config['convergence_relative_tolerance'] = 0.1
        builder = SosModelBuilder()
        builder.construct(config)
        sos_model = builder.finish()
        assert sos_model.convergence_relative_tolerance == 0.1


class TestSosModelBuilder():
    """Tests that the correct SosModel structure is created from a configuration
    dictionary::

        {
            'name': 'sos_model_name',
            'description': 'friendly description of the sos model',
            'sector_models': list of Model,
            'scenario_sets': list of ScenarioModel,
            'max_iterations': int,
            'convergence_absolute_tolerance': float,
            'convergence_relative_tolerance': float,
            'dependencies': [
                {
                    'source_model': str (Model.name),
                    'source_model_output': str (Metadata.name),
                    'sink_model': str (Model.name),
                    'sink_model_output': str (Metadata.name)
                }
            ]
        }

    """
    def test_scenarios(self, get_sos_model_config):
        """Test constructing from single dict config
        """
        config = get_sos_model_config
        builder = SosModelBuilder()
        builder.construct(config)
        sos_model = builder.finish()

        assert isinstance(sos_model, SosModel)
        assert isinstance(sos_model.models['test_scenario_model'], ScenarioModel)

    def test_simple_dependency(self, get_sos_model_config_with_dep):

        config_data = get_sos_model_config_with_dep

        builder = SosModelBuilder()
        builder.construct(config_data)
        sos_model = builder.finish()

        sos_model.make_dependency_graph()
        graph = sos_model.dependency_graph

        scenario = sos_model.models['test_scenario_model']
        model = sos_model.models['water_supply']

        assert 'water_supply' in sos_model.models
        assert sos_model.models['water_supply'] in graph.nodes()
        deps = sos_model.models['water_supply'].deps
        assert 'raininess' in deps.keys()
        expected = Dependency(
            scenario,
            scenario.outputs['raininess'],
            model.inputs['raininess']
        )
        assert deps['raininess'] == expected

        assert 'test_scenario_model' in sos_model.models
        assert sos_model.models['test_scenario_model'] in graph.nodes()

    def test_data_not_present(self, get_sos_model_config_with_dep):
        """Raise a NotImplementedError if an input is defined but no dependency links
        it to a data source
        """
        config_data = get_sos_model_config_with_dep
        config_data['scenario_sets'] = []
        config_data['dependencies'] = []
        with raises(NotImplementedError):
            builder = SosModelBuilder()
            builder.construct(config_data)
            builder.finish()

    def test_undefined_unit_conversion(self, get_sos_model_config_with_dep):

        config_data = get_sos_model_config_with_dep
        sector_model = config_data['sector_models'][0]
        sector_model.inputs['raininess'].units = 'incompatible'

        with raises(ValueError) as ex:
            builder = SosModelBuilder()
            builder.construct(config_data)
            builder.finish()

        assert "Cannot convert to undefined unit 'incompatible'" in str(ex.value)

    def test_invalid_unit_conversion(self, get_sos_model_config_with_dep):

        config_data = get_sos_model_config_with_dep
        scenario = config_data['scenario_sets'][0]

        scenario.outputs['raininess'].units = 'meter'

        with raises(ValueError) as ex:
            builder = SosModelBuilder()
            builder.construct(config_data)
            builder.finish()

        assert "Cannot convert from meter to milliliter" in str(ex.value)

    def test_cyclic_dependencies(self, get_sos_model_config_with_summed_dependency):
        config_data = get_sos_model_config_with_summed_dependency

        builder = SosModelBuilder()

        with raises(NotImplementedError):
            builder.construct(config_data)
        # sos_model = builder.finish()
        # sos_model.check_dependencies()
        # graph = sos_model.dependency_graph

        # scenario = sos_model.models['raininess']
        # water_one = sos_model.models['water_supply']
        # water_two = sos_model.models['water_supply_two']
        # water_three = sos_model.models['water_supply_three']

        # print(graph.edges())

        # assert (scenario, water_one) in graph.edges()
        # assert (scenario, water_two) in graph.edges()
        # assert (water_one, water_three) in graph.edges()
        # assert (water_two, water_three) in graph.edges()



# -*- coding: utf-8 -*-
"""Test metadata access
"""
from pytest import fixture, raises
from smif.convert.area import RegionSet
from smif.convert.interval import IntervalSet
from smif.metadata import Metadata, MetadataSet


@fixture(scope='function')
def two_output_metrics(interval_set, region_set):
    """Returns a model output dictionary with two metrics
    """
    outputs = [
        {
            'name': 'total_cost',
            'spatial_resolution': region_set,
            'temporal_resolution': interval_set,
            'units': 'count'
        },
        {
            'name': 'water_demand',
            'spatial_resolution': region_set,
            'temporal_resolution': interval_set,
            'units': 'count'
        }
    ]
    return outputs


@fixture(scope='function')
def region_set():
    """Return a region set of two square regions::

        |```|```|
        | a | b |
        |...|...|

    """
    rset = RegionSet('half_squares', [
        {
            'type': 'Feature',
            'properties': {'name': 'a'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 1], [1, 1], [1, 0]]]
            }
        },
        {
            'type': 'Feature',
            'properties': {'name': 'b'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 1], [0, 2], [1, 2], [1, 1]]]
            }
        },
    ])
    return rset


@fixture(scope='function')
def interval_set():
    """Return an interval set of the four seasons
    """
    seasons = [
        ('winter', [('P0M', 'P2M'), ('P11M', 'P12M')]),
        ('spring', [('P2M', 'P5M')]),
        ('summer', [('P5M', 'P8M')]),
        ('autumn', [('P8M', 'P11M')])
    ]

    return IntervalSet('seasons', seasons)


class TestMetadata(object):
    """Test Metadata objects
    """
    def test_create_metadata(self, interval_set, region_set):
        """Create Metadata to hold name, spatial and temporal resolution, and units
        """
        metadata = Metadata("total_lane_kilometres", region_set, interval_set,
                            "kilometer")
        assert metadata.name == "total_lane_kilometres"
        assert metadata.spatial_resolution == region_set
        assert metadata.temporal_resolution == interval_set
        assert metadata.units == "kilometer"

    def test_serialise_metadata(self, interval_set, region_set):
        """Create Metadata to hold name, spatial and temporal resolution, and units
        """
        metadata = Metadata("total_lane_kilometres", region_set, interval_set,
                            "kilometer")
        actual = metadata.as_dict()
        expected = {'name': 'total_lane_kilometres',
                    'spatial_resolution': 'half_squares',
                    'temporal_resolution': 'seasons',
                    'units': 'kilometer'}
        assert actual == expected

    def test_metadata_equality(self):
        """Metadata with same attributes should compare equal
        """
        one_m = Metadata("total_lane_kilometres", region_set, interval_set, "kilometer")
        other = Metadata("total_lane_kilometres", region_set, interval_set, "kilometer")
        assert one_m == other

    def test_unit_normalisation(self):
        """Expect units to be set to full names from abbreviation
        """
        metadata = Metadata("total_lane_kilometres", region_set, interval_set,
                            "km")
        assert metadata.units == "kilometer"

    def test_unit_unparseable(self):
        """Expect unrecognised units to be passed through unchanged
        """
        metadata = Metadata("total_lane_kilometres", region_set, interval_set,
                            "unparseable")
        assert metadata.units == "unparseable"

    def test_access_region_names_with_register(self, region_set, interval_set):
        """Metadata should expose region names when a register is available
        """
        rreg = region_set
        metadata = Metadata("total_lane_kilometres", region_set, interval_set, "unparseable")
        metadata.region_register = rreg
        assert metadata.get_region_names() == ["a", "b"]

    def test_access_interval_names(self, interval_set, region_set):
        """Metadata should expose interval names when a register is available
        """
        metadata = Metadata("total_lane_kilometres", region_set, interval_set,
                            "unparseable")
        assert metadata.get_interval_names() == ["winter", "spring", "summer", "autumn"]


class TestMetadataSet(object):
    """Test MetadataSet objects
    """
    def test_create_metadata_set(self, interval_set, region_set):
        """Create MetadataSet to hold a list of Metadata
        """
        metadata_list = [{
            "name": "heat_demand",
            "spatial_resolution": region_set,
            "temporal_resolution": interval_set,
            "units": "kilowatt"
        }]
        metadata_set = MetadataSet(metadata_list)

        # direct access to single metadata
        metadata = metadata_set["heat_demand"]
        assert metadata.name == "heat_demand"
        assert metadata.spatial_resolution == region_set
        assert metadata.temporal_resolution == interval_set
        assert metadata.units == "kilowatt"

        # direct access to list of contained metadata
        assert len(metadata_set) == 1
        assert metadata_set.metadata == [
            Metadata("heat_demand", region_set, interval_set, "kilowatt")
        ]

        # access single metadata attribute
        assert metadata_set.get_spatial_res("heat_demand") == region_set
        assert metadata_set.get_temporal_res("heat_demand") == interval_set
        assert metadata_set.get_units("heat_demand") == "kilowatt"

        # access list of metadata attributes
        assert metadata_set.names == ["heat_demand"]
        assert metadata_set.spatial_resolutions == [region_set]
        assert metadata_set.temporal_resolutions == [interval_set]
        assert metadata_set.units == ["kilowatt"]

    def test_get_spatial_property(self, two_output_metrics, region_set):
        """Access different spatial resolutions for different metadata
        """
        outputs = MetadataSet(two_output_metrics)
        assert outputs.get_spatial_res('total_cost') == region_set
        assert outputs.get_spatial_res('water_demand') == region_set

        with raises(KeyError) as ex:
            outputs.get_spatial_res('missing')
        assert "No metadata found for name 'missing'" in str(ex.value)

    def test_get_temporal_property(self, two_output_metrics, interval_set):
        """Access different temporal resolutions for different metadata
        """
        outputs = MetadataSet(two_output_metrics)
        assert outputs.get_temporal_res('total_cost') == interval_set
        assert outputs.get_temporal_res('water_demand') == interval_set

        with raises(KeyError) as ex:
            outputs.get_temporal_res('missing')
        assert "No metadata found for name 'missing'" in str(ex.value)

    def test_key_error(self):
        """Expect a KeyError on trying to access missing Metadata
        """
        metadata_set = MetadataSet([])
        with raises(KeyError) as ex:
            metadata_set["missing"]
        assert "No metadata found for name 'missing'" in str(ex.value)

    def test_metadata_order(self, interval_set, region_set):
        """Expect list of Metadata to be sorted by name
        """
        metadata_list = [
            {
                "name": "total_lane_kilometres",
                "spatial_resolution": region_set,
                "temporal_resolution": interval_set,
                "units": "kilometer"
            },
            {
                "name": "heat_demand",
                "spatial_resolution": region_set,
                "temporal_resolution": interval_set,
                "units": "kilowatt"
            },
        ]
        metadata_set = MetadataSet(metadata_list)

        assert len(metadata_set) == 2
        assert metadata_set.metadata == [
            Metadata("heat_demand", region_set, interval_set, "kilowatt"),
            Metadata("total_lane_kilometres", region_set, interval_set, "kilometer")
        ]

    def test_access_registers(self, region_set, interval_set):
        """Individual Metadata in a Set should provide region and interval names
        when registers are available
        """
        metadata_list = [
            {
                "name": "total_lane_kilometres",
                "spatial_resolution": region_set,
                "temporal_resolution": interval_set,
                "units": "kilometers"
            }
        ]
        metadata_set = MetadataSet(metadata_list)
        metadata = metadata_set["total_lane_kilometres"]

        assert metadata.get_region_names() == ["a", "b"]
        assert metadata.get_interval_names() == ["winter", "spring", "summer", "autumn"]

    def test_iterate_over_empty(self):
        """Should initialise as empty by default
        """
        metadata_set = MetadataSet()
        assert metadata_set.metadata == []
        assert [x for x in metadata_set] == []

    def test_iterate_over_populated(self):
        """Should initialise with list of metadata if provided
        """
        metadata_list = [
            {
                "name": "total_lane_kilometres",
                "spatial_resolution": region_set,
                "temporal_resolution": interval_set,
                "units": "kilometers"
            }
        ]
        metadata_set = MetadataSet(metadata_list)
        expected_metadata = Metadata(
            "total_lane_kilometres",
            region_set,
            interval_set,
            "kilometers"
        )

        actual = metadata_set.metadata
        assert actual == [expected_metadata]

        actual = [(k, v) for k, v in metadata_set.items()]
        assert actual == [("total_lane_kilometres", expected_metadata)]

    def test_add_meta_object(self):
        """Should allow adding a Metadata object
        """
        metadata = Metadata("total_lane_kilometres", region_set, interval_set,
                            "kilometer")
        metadata_set = MetadataSet()
        metadata_set.add_metadata(metadata)

        # access list of metadata attributes
        assert metadata_set.names == ["total_lane_kilometres"]
        assert metadata_set.spatial_resolutions == [region_set]
        assert metadata_set.temporal_resolutions == [interval_set]
        assert metadata_set.units == ["kilometer"]

    def test_add_meta_dict(self):
        """Should allow adding a dict with required keys
        """
        metadata = {
            "name": "total_lane_kilometres",
            "spatial_resolution": region_set,
            "temporal_resolution": interval_set,
            "units": "kilometers"
        }
        metadata_set = MetadataSet()
        metadata_set.add_metadata(metadata)

        # access list of metadata attributes
        assert metadata_set.names == ["total_lane_kilometres"]
        assert metadata_set.spatial_resolutions == [region_set]
        assert metadata_set.temporal_resolutions == [interval_set]
        assert metadata_set.units == ["kilometer"]



"""Test aggregation/disaggregation of data between sets of areas
"""
import numpy as np
from pytest import fixture, raises
from smif.convert.area import RegionRegister, RegionSet, get_register


@fixture(scope='function')
def regions():
    """Return data structure for test regions/shapes
    """
    return [
        {
            'type': 'Feature',
            'properties': {'name': 'unit'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 1], [1, 1], [1, 0]]]
            }
        },
        {
            'type': 'Feature',
            'properties': {'name': 'half'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 0.5], [1, 0.5], [1, 0]]]
            }
        },
        {
            'type': 'Feature',
            'properties': {'name': 'two'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 2], [1, 2], [1, 0]]]
            }
        }
    ]


@fixture(scope='function')
def regions_half_squares():
    """Return two adjacent square regions::

        |```|```|
        | A | B |
        |...|...|

    """
    return RegionSet('half_squares', [
        {
            'type': 'Feature',
            'properties': {'name': 'a'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 1], [1, 1], [1, 0]]]
            }
        },
        {
            'type': 'Feature',
            'properties': {'name': 'b'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 1], [0, 2], [1, 2], [1, 1]]]
            }
        },
    ])


@fixture(scope='function')
def regions_single_half_square():
    """Return single half-size square region::

        |```|
        | A |
        |...|

    """
    return RegionSet('single_half_square', [
        {
            'type': 'Feature',
            'properties': {'name': 'a'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 1], [1, 1], [1, 0]]]
            }
        }
    ])


@fixture(scope='function')
def regions_rect():
    """Return single region covering 2x1 area::

        |```````|
        |   0   |
        |.......|

    """
    return RegionSet('rect', [
        {
            'type': 'Feature',
            'properties': {'name': 'zero'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 2], [1, 2], [1, 0]]]
            }
        }
    ])


@fixture(scope='function')
def regions_half_triangles():
    """Return regions split diagonally::

        |``````/|
        | 0 / 1 |
        |/......|

    """
    return RegionSet('half_triangles', [
        {
            'type': 'Feature',
            'properties': {'name': 'zero'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [0, 2], [1, 0]]]
            }
        },
        {
            'type': 'Feature',
            'properties': {'name': 'one'},
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[0, 2], [1, 2], [1, 0]]]
            }
        },
    ])


def test_proportion(regions):
    """Sense-check proportion calculator
    """
    region_set = RegionSet('regions', regions)

    assert region_set.get_proportion(0,
                                     region_set[0]) == 1

    assert region_set.get_proportion(0, region_set[1]) == 0.5
    assert region_set.get_proportion(1, region_set[0]) == 1

    assert region_set.get_proportion(0, region_set[2]) == 1
    assert region_set.get_proportion(2, region_set[0]) == 0.5
    assert region_set.get_proportion(1, region_set[2]) == 1
    assert region_set.get_proportion(2, region_set[1]) == 0.25


class TestRegionSet():
    """Test creating, looking up, retrieving region sets
    """

    def test_serialise(self, regions):
        rset = RegionSet('test', regions)
        rset.description = 'my description'
        actual = rset.as_dict()
        expected = {'name': 'test',
                    'description': 'my description'}

        assert actual == expected

    def test_create(self, regions):
        rset = RegionSet('test', regions)
        assert rset.name == 'test'
        assert len(rset) == 3
        assert rset[0].name == 'unit'
        assert rset[1].name == 'half'
        assert rset[2].name == 'two'

    def test_get_names(self, regions):
        rset = RegionSet('test', regions)
        actual = rset.get_entry_names()
        expected = ['unit', 'half', 'two']
        assert actual == expected

    def test_as_features(self, regions_half_squares):
        """Retrieve regions as feature dicts
        """
        actual = regions_half_squares.as_features()
        expected = [
            {
                'type': 'Feature',
                'properties': {'name': 'a'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': (((0.0, 0.0), (0.0, 1.0), (1.0, 1.0),
                                     (1.0, 0.0), (0.0, 0.0),),)
                }
            },
            {
                'type': 'Feature',
                'properties': {'name': 'b'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': (((0.0, 1.0), (0.0, 2.0), (1.0, 2.0),
                                     (1.0, 1.0), (0.0, 1.0),),)
                }
            },
        ]
        assert actual == expected

    def test_centroids_as_features(self, regions_half_squares):
        """Retrieve centroids
        """
        actual = regions_half_squares.centroids_as_features()
        expected = [
            {
                'type': 'Feature',
                'properties': {'name': 'a'},
                'geometry': {
                    'type': 'Point',
                    'coordinates': (0.5, 0.5)
                }
            },
            {
                'type': 'Feature',
                'properties': {'name': 'b'},
                'geometry': {
                    'type': 'Point',
                    'coordinates': (0.5, 1.5)
                }
            },
        ]
        assert actual == expected

    def test_must_have_unique_names(self):
        with raises(AssertionError) as ex:
            RegionSet('test', [
                {
                    'type': 'Feature',
                    'properties': {'name': 'a'},
                    'geometry': {
                        'type': 'Polygon',
                        'coordinates': [[[0, 0], [0, 1], [1, 1], [1, 0]]]
                    }
                },
                {
                    'type': 'Feature',
                    'properties': {'name': 'a'},
                    'geometry': {
                        'type': 'Polygon',
                        'coordinates': [[[0, 1], [0, 2], [1, 2], [1, 1]]]
                    }
                },
            ])

        assert 'Region set must have uniquely named regions' in str(ex)


class TestRegionRegister():
    """Test creating registers, registering region sets, converting data
    """
    def test_create(self):
        rreg = RegionRegister()
        assert rreg.names == []

        with raises(ValueError) as ex:
            rreg.get_entry('nonexistent')
        assert "ResolutionSet 'nonexistent' not registered" in str(ex)

    def test_convert_equal(self):
        rreg = get_register()

        data = np.ones(1)
        converted = rreg.convert(data, 'rect', 'rect_alt')
        np.testing.assert_equal(data, converted)

    def test_convert_to_half(self):
        rreg = get_register()

        data = np.ones(1)
        converted = rreg.convert(data, 'rect', 'half_squares')
        expected = np.array([0.5, 0.5])
        np.testing.assert_equal(converted, expected)

    def test_convert_from_half(self):
        rreg = get_register()

        data = np.ones(2) / 2
        converted = rreg.convert(data, 'half_squares', 'rect')
        expected = np.ones(1)
        np.testing.assert_equal(converted, expected)

        data = np.array([2, 3])
        converted = rreg.convert(data, 'half_squares', 'rect')
        expected = np.array([5])
        np.testing.assert_equal(converted, expected)

    def test_coverage_half(self):
        rreg = get_register()
        half_covered = rreg.get_entry('single_half_square')
        actual = half_covered.coverage
        expected = 1.0
        assert actual == expected

    def test_coverage_whole(self):
        rreg = get_register()
        rect = rreg.get_entry('rect')
        actual = rect.coverage
        expected = 2.0
        assert actual == expected

    def test_convert_to_half_not_covered(self, caplog):
        rreg = get_register()

        data = np.array([3])

        rreg.convert(data, 'rect', 'single_half_square')

        expected = "Coverage for 'rect' is 2 and does not match " \
                   "coverage for 'single_half_square' which is 1"

        assert expected in caplog.text

    def test_convert_from_half_not_covered(self, caplog):
        rreg = get_register()

        data = np.array([3])
        rreg.convert(data, 'single_half_square', 'rect')

        expected = "Coverage for 'single_half_square' is 1 and does not " \
                   "match coverage for 'rect' which is 2"

        assert expected in caplog.text

    def test_convert_square_to_triangle(self):
        rreg = get_register()

        data = np.array([1, 1])
        converted = rreg.convert(data, 'half_squares', 'half_triangles')
        expected = np.array([1, 1])
        np.testing.assert_equal(converted, expected)

        data = np.array([0, 1])
        converted = rreg.convert(data, 'half_squares', 'half_triangles')
        expected = np.array([0.25, 0.75])
        np.testing.assert_equal(converted, expected)

    def test_convert_triangle_to_square(self):
        rreg = get_register()

        data = np.array([1, 1])
        converted = rreg.convert(data, 'half_triangles', 'half_squares')
        expected = np.array([1, 1])
        np.testing.assert_equal(converted, expected)

        data = np.array([0.25, 0.75])
        converted = rreg.convert(data, 'half_triangles', 'half_squares')
        expected = np.array([0.375, 0.625])
        np.testing.assert_equal(converted, expected)


class TestGetCoefficients:

    def test_get_coefficients(self):
        rreg = get_register()
        actual = rreg.get_coefficients('half_triangles',
                                       'half_squares')
        expected = np.array([[0.75, 0.25],
                             [0.25, 0.75]])
        np.testing.assert_equal(actual, expected)

    def test_coefs_should_map_to_themselves(self):
        rreg = get_register()
        actual = rreg.get_coefficients('half_triangles',
                                       'half_triangles')
        expected = np.array([[1, 0],
                             [0, 1]])
        np.testing.assert_equal(actual, expected)



# -*- coding: utf-8 -*-
"""Validate the correct format and presence of the config data
for the system-of-systems model
"""

VALIDATION_ERRORS = []


class ValidationError(Exception):
    """Custom exception to use for parsing validation.
    """
    pass


def validate_sos_model_config(data):
    """Check expected values for data loaded from master config file
    """
    if not isinstance(data, dict):
        msg = "Main config file should contain setup data, instead found: {}"
        err = ValidationError(msg.format(data))
        VALIDATION_ERRORS.append(err)
        return

    # check dependencies
    if "dependencies" not in data:
        VALIDATION_ERRORS.append(
            ValidationError("No 'dependencies' specified in main config file.")
        )

    # check timesteps
    if "timesteps" not in data:
        VALIDATION_ERRORS.append(
            ValidationError("No 'timesteps' file specified in main config file."))
    else:
        validate_path_to_timesteps(data["timesteps"])

    # check sector models
    if "sector_models" not in data:
        VALIDATION_ERRORS.append(
            ValidationError("No 'sector_models' specified in main config file."))
    else:
        validate_sector_models_initial_config(data["sector_models"])

    # check scenario data
    if "scenario_data" not in data:
        VALIDATION_ERRORS.append(
            ValidationError("No 'scenario_data' specified in main config file."))
    else:
        validate_scenario_data_config(data["scenario_data"])

    # check planning
    if "planning" not in data:
        VALIDATION_ERRORS.append(
            ValidationError("No 'planning' mode specified in main config file."))
    else:
        validate_planning_config(data["planning"])

    # check region_sets
    if "region_sets" not in data:
        VALIDATION_ERRORS.append(
            ValidationError("No 'region_sets' specified in main config file."))
    else:
        validate_region_sets_config(data["region_sets"])

    # check interval_sets
    if "interval_sets" not in data:
        VALIDATION_ERRORS.append(
            ValidationError("No 'interval_sets' specified in main config file."))
    else:
        validate_interval_sets_config(data["interval_sets"])


def validate_path_to_timesteps(timesteps):
    """Check timesteps is a path to timesteps file
    """
    if not isinstance(timesteps, str):
        VALIDATION_ERRORS.append(
            ValidationError(
                "Expected 'timesteps' in main config to specify " +
                "a timesteps file, instead got {}.".format(timesteps)))


def validate_timesteps(timesteps, file_path):
    """Check timesteps is a list of integers
    """
    if not isinstance(timesteps, list):
        msg = "Loading {}: expected a list of timesteps.".format(file_path)
        VALIDATION_ERRORS.append(ValidationError(msg))
    else:
        msg = "Loading {}: timesteps should be integer years, instead got {}"
        for timestep in timesteps:
            if not isinstance(timestep, int):
                VALIDATION_ERRORS.append(msg.format(file_path, timestep))


def validate_time_intervals(intervals, file_path):
    """Check time intervals
    """
    if not isinstance(intervals, list):
        msg = "Loading {}: expected a list of time intervals.".format(file_path)
        VALIDATION_ERRORS.append(ValidationError(msg))
    else:
        for interval in intervals:
            validate_time_interval(interval)


def validate_time_interval(interval):
    """Check a single time interval
    """
    if not isinstance(interval, dict):
        msg = "Expected a time interval, instead got {}.".format(interval)
        VALIDATION_ERRORS.append(ValidationError(msg))
        return

    required_keys = ["id", "start", "end"]
    for key in required_keys:
        if key not in interval:
            fmt = "Expected a value for '{}' in each " + \
                "time interval, only received {}"
            VALIDATION_ERRORS.append(ValidationError(fmt.format(key, interval)))


def validate_sector_models_initial_config(sector_models):
    """Check list of sector models initial configuration
    """
    if not isinstance(sector_models, list):
        fmt = "Expected 'sector_models' in main config to " + \
              "specify a list of sector models to run, instead got {}."
        VALIDATION_ERRORS.append(ValidationError(fmt.format(sector_models)))
    else:
        if len(sector_models) == 0:
            VALIDATION_ERRORS.append(
                ValidationError("No 'sector_models' specified in main config file."))

        # check each sector model
        for sector_model_config in sector_models:
            validate_sector_model_initial_config(sector_model_config)


def validate_sector_model_initial_config(sector_model_config):
    """Check a single sector model initial configuration
    """
    if not isinstance(sector_model_config, dict):
        fmt = "Expected a sector model config block, instead got {}"
        VALIDATION_ERRORS.append(ValidationError(fmt.format(sector_model_config)))
        return

    required_keys = ["name", "config_dir", "path", "classname"]
    for key in required_keys:
        if key not in sector_model_config:
            fmt = "Expected a value for '{}' in each " + \
                  "sector model in main config file, only received {}"
            VALIDATION_ERRORS.append(ValidationError(fmt.format(key, sector_model_config)))


def validate_dependency_spec(input_spec, model_name):
    """Check the input specification for a single sector model
    """
    if not isinstance(input_spec, list):
        fmt = "Expected a list of parameter definitions in '{}' model " + \
              "input specification, instead got {}"
        VALIDATION_ERRORS.append(ValidationError(fmt.format(model_name, input_spec)))
        return

    for dep in input_spec:
        validate_dependency(dep)


def validate_dependency(dep):
    """Check a dependency specification
    """
    if not isinstance(dep, dict):
        fmt = "Expected a dependency specification, instead got {}"
        VALIDATION_ERRORS.append(ValidationError(fmt.format(dep)))
        return

    required_keys = ["name", "spatial_resolution", "temporal_resolution", "units"]
    for key in required_keys:
        if key not in dep:
            fmt = "Expected a value for '{}' in each model dependency, only received {}"
            VALIDATION_ERRORS.append(ValidationError(fmt.format(key, dep)))


def validate_scenario_data_config(scenario_data):
    """Check scenario data
    """
    if not isinstance(scenario_data, list):
        fmt = "Expected a list of scenario datasets in main model config, " + \
              "instead got {}"
        VALIDATION_ERRORS.append(ValidationError(fmt.format(scenario_data)))
        return

    for scenario in scenario_data:
        validate_scenario(scenario)


def validate_scenario(scenario):
    """Check a single scenario specification
    """
    if not isinstance(scenario, dict):
        fmt = "Expected a scenario specification, instead got {}"
        VALIDATION_ERRORS.append(ValidationError(fmt.format(scenario)))
        return

    required_keys = ["parameter", "spatial_resolution", "temporal_resolution", "units", "file"]
    for key in required_keys:
        if key not in scenario:
            fmt = "Expected a value for '{}' in each scenario, only received {}"
            VALIDATION_ERRORS.append(ValidationError(fmt.format(key, scenario)))


def validate_scenario_data(data, file_path):
    """Check a list of scenario observations
    """
    if not isinstance(data, list):
        fmt = "Expected a list of scenario data in {}"
        VALIDATION_ERRORS.append(ValidationError(fmt.format(file_path)))
        return

    for datum in data:
        validate_scenario_datum(datum, file_path)


def validate_scenario_datum(datum, file_path):
    """Check a single scenario datum
    """
    if not isinstance(datum, dict):
        fmt = "Expected a scenario data point, instead got {}"
        VALIDATION_ERRORS.append(ValidationError(fmt.format(datum)))
        return

    required_keys = ["region", "interval", "year", "value"]
    for key in required_keys:
        if key not in datum:
            fmt = "Expected a value for '{}' in each data point in a scenario, " + \
                  "only received {}"
            VALIDATION_ERRORS.append(ValidationError(fmt.format(key, datum)))


def validate_initial_conditions(data, file_path):
    """Check a list of initial condition observations
    """
    if not isinstance(data, list):
        fmt = "Expected a list of initial conditions in {}"
        VALIDATION_ERRORS.append(ValidationError(fmt.format(file_path)))
        return

    for datum in data:
        validate_initial_condition(datum, file_path)


def validate_initial_condition(datum, file_path):
    """Check a single initial condition datum
    """
    if not isinstance(datum, dict):
        fmt = "Expected a initial condition data point, instead got {} from {}"
        VALIDATION_ERRORS.append(ValidationError(fmt.format(datum, file_path)))
        return

    required_keys = ["name", "location", "capital_cost", "operational_lifetime",
                     "economic_lifetime"]
    for key in required_keys:
        if key not in datum:
            fmt = "Expected a value for '{}' in each data point in a initial condition, " + \
                  "only received {} from {}"
            VALIDATION_ERRORS.append(ValidationError(fmt.format(key, datum, file_path)))


def validate_planning_config(planning):
    """Check planning options
    """
    required_keys = ["pre_specified", "rule_based", "optimisation"]
    for key in required_keys:
        if key not in planning:
            fmt = "No '{}' settings specified under 'planning' " + \
                  "in main config file."
            VALIDATION_ERRORS.append(ValidationError(fmt.format(key)))

    # check each planning type
    for key, planning_type in planning.items():
        if "use" not in planning_type:
            fmt = "No 'use' settings specified for '{}' 'planning'"
            VALIDATION_ERRORS.append(ValidationError(fmt.format(key)))
            continue
        if planning_type["use"]:
            if "files" not in planning_type or \
               not isinstance(planning_type["files"], list) or \
               len(planning_type["files"]) == 0:

                fmt = "No 'files' provided for the '{}' " + \
                      "planning type in main config file."
                VALIDATION_ERRORS.append(ValidationError(fmt.format(key)))


def validate_region_sets_config(region_sets):
    """Check regions sets
    """
    required_keys = ["name", "file"]
    for key in required_keys:
        for region_set in region_sets:
            if key not in region_set:
                fmt = "Expected a value for '{}' in each " + \
                    "region set in main config file, only received {}"
                VALIDATION_ERRORS.append(ValidationError(fmt.format(key, region_set)))


def validate_interval_sets_config(interval_sets):
    """Check interval sets
    """
    required_keys = ["name", "file"]
    for key in required_keys:
        for interval_set in interval_sets:
            if key not in interval_set:
                fmt = "Expected a value for '{}' in each " + \
                    "interval set in main config file, only received {}"
                VALIDATION_ERRORS.append(ValidationError(fmt.format(key, interval_set)))


def validate_interventions(data, path):
    """Validate the loaded data as required for model interventions
    """
    # check required keys
    required_keys = ["name", "location", "capital_cost", "operational_lifetime",
                     "economic_lifetime"]

    # except for some keys which are allowed simple values,
    # expect each attribute to be of the form {value: x, units: y}
    simple_keys = ["name", "sector", "location"]

    for intervention in data:
        for key in required_keys:
            if key not in intervention:
                fmt = "Loading interventions from {}, required " + \
                      "a value for '{}' in each intervention, but only " + \
                      "received {}"
                VALIDATION_ERRORS.append(ValidationError(fmt.format(path, key, intervention)))

        for key, value in intervention.items():
            if key not in simple_keys and (
                    not isinstance(value, dict)
                    or "value" not in value
                    or "units" not in value):
                fmt = "Loading interventions from {3}, {0}.{1} was {2} but " + \
                      "should have specified units, " + \
                      "e.g. {{'value': {2}, 'units': 'm'}}"

                msg = fmt.format(intervention["name"], key, value, path)
                VALIDATION_ERRORS.append(ValidationError(msg))



"""Test config validation
"""
from pytest import fixture
from smif.data_layer.validate import (VALIDATION_ERRORS, validate_dependency,
                                      validate_dependency_spec,
                                      validate_initial_conditions,
                                      validate_interval_sets_config,
                                      validate_interventions,
                                      validate_planning_config,
                                      validate_region_sets_config,
                                      validate_scenario,
                                      validate_scenario_data,
                                      validate_scenario_data_config,
                                      validate_scenario_datum,
                                      validate_sector_model_initial_config,
                                      validate_sos_model_config,
                                      validate_time_intervals,
                                      validate_timesteps)


@fixture(scope='function')
def get_sos_model_config():
    """Return sample sos_model config
    """
    return {
        "sector_models": [
            {
                "name": "water_supply",
                "path": "../models/water_supply.py",
                "classname": "WaterSupplySectorModel",
                "config_dir": "../data/water_supply",
                "initial_conditions": [
                    "../data/water_supply/initial_conditions/initial_2015_oxford.yaml"
                ],
                "interventions": [
                    "../data/water_supply/interventions/types.yaml"
                ]
            }
        ],
        "timesteps": "timesteps.yaml",
        "scenario_data": [
            {
                "parameter": "population",
                "file": "../data/scenario/population.yaml",
                'spatial_resolution': 'national',
                'temporal_resolution': 'annual',
                'units': 'count'
            },
            {
                "parameter": "raininess",
                "file": "../data/scenario/raininess.yaml",
                'spatial_resolution': 'national',
                'temporal_resolution': 'annual',
                'units': 'ml'
            }
        ],
        "region_sets": [{'name': 'national',
                         'file': '../data/national.shp'}],
        "interval_sets": [{'name': 'annual',
                           'file': '../data/annual_intervals.yaml'}],
        "planning": {
            "pre_specified": {
                "use": True,
                "files": [
                    "../planning/expected_to_2020.yaml",
                    "../planning/national_infrastructure_pipeline.yaml"
                ]
            },
            "rule_based": {
                "use": False
            },
            "optimisation": {
                "use": False
            }
        }
    }


@fixture(scope='function')
def get_sector_model_initial_config():
    """Return sample sector_model initial config
    """
    return {
        'name': 'test_model',
        'config_dir': '/path/to/config',
        'path': '/path/to/model/run.py',
        'classname': 'SectorModelImplementation'
    }


@fixture(scope='function')
def get_intervention():
    """Return sample intervention
    """
    return {
        'name': 'asset',
        'capacity': {
            'value': 3,
            'units': 'MW'
        },
        'location': 'galloway',
        'operational_lifetime': {
            'value': 150,
            'units': "years"
        },
        'economic_lifetime': {
            'value': 50,
            'units': "years"
        },
        'capital_cost': {
            'value': 50,
            'units': "million £/km"
        }
    }


@fixture(scope='function')
def get_scenario_data():
    """Return sample scenario data
    """
    return [
        {
            "region": "England",
            "interval": 1,
            "year": 2010,
            "value": 52000000,
            "units": "people",
        },
        {
            "region": "Scotland",
            "interval": 1,
            "year": 2010,
            "value": 5100000,
            "units": "people",
        },
        {
            "region": "Wales",
            "interval": 1,
            "year": 2010,
            "value": 2900000,
            "units": "people",
        },
    ]


@fixture(scope='function')
def get_initial_condition():
    """Return sample initial condition (pre-existing intervention)
    """
    return {
        "name": "test",
        "location": "UK",
        "capital_cost": {
            "value": 150,
            "units": "million GBP"
        },
        "operational_lifetime": {
            "value": 150,
            "units": "years"
        },
        "economic_lifetime": {
            "value": 150,
            "units": "years"
        },
        "build_date": 1975
    }


@fixture(scope='function')
def get_dependency_spec():
    """Return sample input specification
    """
    return [
        {
            'name': 'gas_demand',
            'spatial_resolution': 'national',
            'temporal_resolution': 'annual',
            'units': 'Ml'
        }
    ]


@fixture(scope='function')
def get_dependency():
    """Return sample dependency
    """
    return {
        'name': 'gas_demand',
        'spatial_resolution': 'national',
        'temporal_resolution': 'annual',
        'units': 'Ml'
    }


@fixture(scope='function')
def get_time_intervals():
    """Return sample time intervals
    """
    return [
        {
            'id': 'first_half',
            'start': 'P0M',
            'end': 'P6M'
        },
        {
            'id': 'second_half',
            'start': 'P6M',
            'end': 'P12M'
        }
    ]


def test_modelrun_config_invalid():
    """Expect an error if not a dict
    """
    invalid_possibilities = [
        0,
        [],
        "just a string",
        3.1415
    ]

    for invalid_data in invalid_possibilities:
        validate_sos_model_config(invalid_data)
        ex = VALIDATION_ERRORS.pop()
        msg = "Main config file should contain setup data"
        assert msg in str(ex)


def test_invalid_timesteps_file(get_sos_model_config):
    """Expect an error if timesteps is not a path to a file
    """
    data = get_sos_model_config
    data['timesteps'] = 3

    validate_sos_model_config(data)
    ex = VALIDATION_ERRORS.pop()
    msg = "Expected 'timesteps' in main config to specify a timesteps file, instead got 3"
    assert msg in str(ex)


def test_invalid_timesteps():
    """Expect a list of timesteps, else error
    """
    data = 2010
    validate_timesteps(data, "timestep.yaml")
    ex = VALIDATION_ERRORS.pop()
    msg = "expected a list of timesteps"
    assert msg in str(ex)


def test_invalid_single_timestep():
    """Expect an error for non-integer timesteps
    """
    data = [2010, "January 2015"]
    validate_timesteps(data, "timestep.yaml")
    ex = VALIDATION_ERRORS.pop()
    msg = "timesteps should be integer years"
    assert msg in str(ex)


def test_invalid_initial_conditions(get_initial_condition):
    """Expect an error if initial conditions is not a list of dicts
    """
    validate_initial_conditions(234, 'test.yaml')
    ex = VALIDATION_ERRORS.pop()
    msg = "Expected a list of initial conditions"
    assert msg in str(ex)

    validate_initial_conditions(['do', 're', 'mi'], 'test.yaml')
    ex = VALIDATION_ERRORS.pop()
    msg = "Expected a initial condition data point"
    assert msg in str(ex)

    required_keys = ["name", "location", "capital_cost", "operational_lifetime",
                     "economic_lifetime"]
    for key in required_keys:
        datum = get_initial_condition
        del datum[key]
        validate_initial_conditions([datum], 'test.yaml')
        msg = "Expected a value for '{}' in each data point in a initial condition".format(key)
        ex = VALIDATION_ERRORS.pop()
        assert msg in str(ex)


def test_time_intervals_type():
    """Expect an error if a set of time_intervals is not a list
    """
    data = 'Jiminy Cricket'
    validate_time_intervals(data, '/path/to/interval_set.yaml')
    ex = VALIDATION_ERRORS.pop()
    msg = "expected a list of time intervals"
    assert msg in str(ex)


def test_time_interval_type():
    """Expect an error if a set of time_intervals is not a list
    """
    data = ['Jiminy Cricket']
    validate_time_intervals(data, '/path/to/interval_set.yaml')
    ex = VALIDATION_ERRORS.pop()
    msg = "Expected a time interval, instead got Jiminy Cricket"
    assert msg in str(ex)


def test_time_interval_required(get_time_intervals):
    """Expect an error if a set of time_intervals is not a list
    """
    required_keys = ['id', 'start', 'end']
    for key in required_keys:
        data = get_time_intervals
        del data[0][key]

        validate_time_intervals(data, '/path/to/interval_set.yaml')
        ex = VALIDATION_ERRORS.pop()
        msg = "Expected a value for '{}' in each time interval".format(key)
        assert msg in str(ex)


def test_missing_sector_models(get_sos_model_config):
    """Expect an error if missing sector_models
    """
    data = get_sos_model_config
    del data['sector_models']

    validate_sos_model_config(data)
    ex = VALIDATION_ERRORS.pop()
    msg = "No 'sector_models' specified in main config"
    assert msg in str(ex)


def test_sector_models_not_list(get_sos_model_config):
    """Expect an error if sector_models is not a list
    """
    data = get_sos_model_config
    data['sector_models'] = 42

    validate_sos_model_config(data)
    ex = VALIDATION_ERRORS.pop()
    msg = "Expected 'sector_models' in main config to specify a list of sector " + \
          "models to run, instead got 42."
    assert msg in str(ex)


def test_sector_models_empty_list(get_sos_model_config):
    """Expect an error if sector_models is an empty list
    """
    data = get_sos_model_config
    data['sector_models'] = []

    validate_sos_model_config(data)
    ex = VALIDATION_ERRORS.pop()
    msg = "No 'sector_models' specified in main config"
    assert msg in str(ex)


def test_sector_model_type(get_sos_model_config):
    """Expect an error if sector_model config is not a dict
    """
    data = get_sos_model_config
    data['sector_models'] = [None]

    validate_sos_model_config(data)
    ex = VALIDATION_ERRORS.pop()
    msg = "Expected a sector model config block"
    assert msg in str(ex)


def test_sector_model_required(get_sector_model_initial_config):
    """Expect an error if a sector_model is missing a required key
    """
    required_keys = ['name', 'config_dir', 'path', 'classname']
    for key in required_keys:
        data = get_sector_model_initial_config
        del data[key]

        validate_sector_model_initial_config(data)
        ex = VALIDATION_ERRORS.pop()
        msg = "Expected a value for '{}'".format(key)
        assert msg in str(ex)


def test_input_spec_type():
    """Expect an error if input_spec is not a dict
    """
    data = None
    validate_dependency_spec(data, 'energy_demand')
    ex = VALIDATION_ERRORS.pop()
    msg = "Expected a list of parameter definitions in 'energy_demand' model input " + \
          "specification, instead got None"
    assert msg in str(ex)


def test_input_spec_deps():
    """Expect an error if input_spec is missing dependencies
    """
    data = {}
    validate_dependency_spec(data, 'energy_demand')
    ex = VALIDATION_ERRORS.pop()
    msg = "Expected a list of parameter definitions in 'energy_demand' model input " + \
          "specification, instead got {}"
    assert msg in str(ex)


def test_input_spec_deps_list():
    """Expect an error if input_spec dependencies is not a list
    """
    data = {1.618}
    validate_dependency_spec(data, 'energy_demand')
    ex = VALIDATION_ERRORS.pop()
    msg = "Expected a list of parameter definitions in 'energy_demand' model input " + \
          "specification, instead got {1.618}"
    assert msg in str(ex)


def test_input_spec_deps_list_empty():
    """Expect no errors for an empty list
    """
    data = []
    # empty is ok
    n_errors_before = len(VALIDATION_ERRORS)
    validate_dependency_spec(data, 'energy_demand')
    assert len(VALIDATION_ERRORS) == n_errors_before


def test_input_spec_deps_list_ok(get_dependency_spec):
    """Expect no errors for a list of valid dependencies
    """
    data = get_dependency_spec
    n_errors_before = len(VALIDATION_ERRORS)
    validate_dependency_spec(data, 'energy_demand')
    assert len(VALIDATION_ERRORS) == n_errors_before


def test_dependency_type(get_dependency_spec):
    """Expect an error if dependency is not a dict
    """
    data = 'single_string'
    validate_dependency(data)
    ex = VALIDATION_ERRORS.pop()
    msg = "Expected a dependency specification, " + \
          "instead got single_string"
    assert msg in str(ex)


def test_dependency_ok(get_dependency):
    """Expect no errors for a valid dependency
    """
    data = get_dependency
    n_errors_before = len(VALIDATION_ERRORS)
    validate_dependency(data)
    assert len(VALIDATION_ERRORS) == n_errors_before


def test_dependency_required(get_dependency):
    """Expect an error if dependency is missing required fields
    """
    required_keys = ['name', 'spatial_resolution', 'temporal_resolution', 'units']
    for key in required_keys:
        data = get_dependency
        del data[key]
        validate_dependency(data)
        ex = VALIDATION_ERRORS.pop()
        msg = "Expected a value for '{}' in each model dependency, " + \
            "only received {}"
        assert msg.format(key, data) in str(ex)


def test_missing_planning(get_sos_model_config):
    """Expect an error if missing planning
    """
    data = get_sos_model_config
    del data['planning']

    validate_sos_model_config(data)
    ex = VALIDATION_ERRORS.pop()
    msg = "No 'planning' mode specified in main config"
    assert msg in str(ex)


def test_used_planning_needs_files(get_sos_model_config):
    """Expect an error if a planning mode is to be used, but has no files
    """
    data = get_sos_model_config
    del data["planning"]["pre_specified"]["files"]

    validate_sos_model_config(data)
    ex = VALIDATION_ERRORS.pop()
    msg = "No 'files' provided for the 'pre_specified' planning type in main config"
    assert msg in str(ex)


def test_planning_missing_required(get_sos_model_config):
    """Expect an error if missing a planning mode
    """
    required_keys = ["pre_specified", "rule_based", "optimisation"]
    for key in required_keys:
        data = get_sos_model_config["planning"]
        del data[key]

        validate_planning_config(data)
        ex = VALIDATION_ERRORS.pop()
        msg = "No '{}' settings specified under 'planning'".format(key)
        assert msg in str(ex)


def test_planning_missing_use(get_sos_model_config):
    """Expect an error if a planning mode is missing "use""
    """
    data = get_sos_model_config["planning"]
    del data["rule_based"]["use"]

    validate_planning_config(data)
    ex = VALIDATION_ERRORS.pop()
    msg = "No 'use' settings specified for 'rule_based' 'planning'"
    assert msg in str(ex)


def test_interventions_missing_required(get_intervention):
    """Expect an error if an intervention is missing required key
    """
    required_keys = ["name", "location", "capital_cost",
                     "operational_lifetime",
                     "economic_lifetime"]

    for key in required_keys:
        intervention = get_intervention
        del intervention[key]

        data = [intervention]

        msg = "required a value for '{}' in each intervention".format(key)

        validate_interventions(data, "/path/to/data.yaml")
        ex = VALIDATION_ERRORS.pop()
        assert msg in str(ex)


def test_interventions_checks_for_units(get_intervention):
    """Expect an error if an intervention's "capacity" has no units
    """
    intervention = get_intervention
    intervention["capacity"] = 3

    data = [intervention]

    msg = "Loading interventions from /path/to/data.yaml, asset.capacity " + \
          "was 3 but should have specified units, e.g. " + \
          "{'value': 3, 'units': 'm'}"

    validate_interventions(data, "/path/to/data.yaml")
    ex = VALIDATION_ERRORS.pop()
    assert msg in str(ex)


def test_scenario_config_type(get_sos_model_config):
    """Expect an error scenario data config is not a list of dicts
    """
    validate_scenario_data_config(42)
    msg = "Expected a list of scenario datasets in main model config"
    ex = VALIDATION_ERRORS.pop()
    assert msg in str(ex)

    validate_scenario_data_config([42])
    msg = "Expected a scenario specification"
    ex = VALIDATION_ERRORS.pop()
    assert msg in str(ex)


def test_scenario_config_missing_required(get_sos_model_config):
    """Expect an error if a scenario datum is missing required key
    """
    required_keys = ["parameter", "spatial_resolution", "temporal_resolution", "units", "file"]

    for key in required_keys:
        data = get_sos_model_config['scenario_data'][0]
        del data[key]

        msg = "Expected a value for '{}' in each scenario".format(key)

        validate_scenario(data)
        ex = VALIDATION_ERRORS.pop()
        assert msg in str(ex)


def test_scenario_missing_required(get_scenario_data):
    """Expect an error if a scenario datum is missing required key
    """
    required_keys = ["region", "interval", "year", "value"]

    for key in required_keys:
        data = get_scenario_data
        for obs in data:
            del obs[key]

        msg = "Expected a value for '{}' in each data point in a scenario".format(key)

        validate_scenario_data(data, "/path/to/data.yaml")
        ex = VALIDATION_ERRORS.pop()
        assert msg in str(ex)


def test_scenario_data_type():
    """Expect an error if scenario data is not a list, or any datum is not a dict
    """
    validate_scenario_data(None, "/path/to/data.yaml")
    ex = VALIDATION_ERRORS.pop()
    assert "Expected a list of scenario data" in str(ex)

    validate_scenario_datum([], "/path/to/data.yaml")
    msg = "Expected a scenario data point, instead got []"
    ex = VALIDATION_ERRORS.pop()
    assert msg in str(ex)


class TestValidateMissingConfigSections:
    """Check that validation raises validation errors when whole sections of
    the configuration data are missing

    """
    def test_modelrun_config_validate(self, get_sos_model_config):
        """Expect no errors from a complete, valid config
        """
        validate_sos_model_config(get_sos_model_config)

    def test_missing_timestep(self, get_sos_model_config):
        """Expect an error if missing timesteps
        """
        data = get_sos_model_config
        del data['timesteps']

        validate_sos_model_config(data)
        ex = VALIDATION_ERRORS.pop()
        msg = "No 'timesteps' file specified in main config"
        assert msg in str(ex)

    def test_region_sets_missing(self, get_sos_model_config):
        """Expect an error if no region sets are specified
        """
        data = get_sos_model_config
        del data['region_sets']

        validate_sos_model_config(data)
        ex = VALIDATION_ERRORS.pop()
        msg = "No 'region_sets' specified in main config file."
        assert msg == str(ex)

    def test_region_sets_required(self):
        missing_name = {'file': 'test.yaml'}
        validate_region_sets_config(missing_name)
        ex = VALIDATION_ERRORS.pop()
        msg = "Expected a value for 'name' in each region set in main config file"
        assert msg in str(ex)

        missing_file = {'name': 'test'}
        validate_region_sets_config(missing_file)
        ex = VALIDATION_ERRORS.pop()
        msg = "Expected a value for 'file' in each region set in main config file"
        assert msg in str(ex)

    def test_interval_sets_missing(self, get_sos_model_config):
        """Expect an error if no time interval sets are specified
        """
        data = get_sos_model_config
        del data['interval_sets']

        validate_sos_model_config(data)
        ex = VALIDATION_ERRORS.pop()
        msg = "No 'interval_sets' specified in main config file."
        assert msg == str(ex)

    def test_interval_sets_required(self):
        missing_name = {'file': 'test.yaml'}
        validate_interval_sets_config(missing_name)
        ex = VALIDATION_ERRORS.pop()
        msg = "Expected a value for 'name' in each interval set in main config file"
        assert msg in str(ex)

        missing_file = {'name': 'test'}
        validate_interval_sets_config(missing_file)
        ex = VALIDATION_ERRORS.pop()
        msg = "Expected a value for 'file' in each interval set in main config file"
        assert msg in str(ex)

    def test_scenario_data_missing(self, get_sos_model_config):
        """Expect an error if no scenario data is specified, but referenced in
        input file
        """
        data = get_sos_model_config
        del data['scenario_data']

        validate_sos_model_config(data)
        ex = VALIDATION_ERRORS.pop()
        msg = "No 'scenario_data' specified in main config file."
        assert msg == str(ex)


class TestValidateMissingConfigurationReferences:
    """Check that validation raises validation errors when one part of the
    configuration assumes presence of a file or other configuration that is
    missing
    """
    def test_scenario_files_missing(self):
        """Expect an error if scenario data files are missing
        """
        pass



import logging
import os
import sys
import traceback

from smif.data_layer import DatafileInterface, DataNotFoundError
from smif.model.scenario_model import ScenarioModelBuilder
from smif.model.sector_model import SectorModelBuilder
from smif.model.sos_model import SosModelBuilder
from smif.modelrun import ModelRunBuilder
from smif.parameters import Narrative

LOGGER = logging.getLogger(__name__)


def get_model_run_definition(directory, modelrun):
    """Builds the model run

    Arguments
    ---------
    directory : str
        Path to the project directory
    modelrun : str
        Name of the model run to run

    Returns
    -------
    dict
        The complete sos_model_run configuration dictionary with contained
        ScenarioModel, SosModel and SectorModel objects

    """
    handler = DatafileInterface(directory)
    try:
        model_run_config = handler.read_sos_model_run(modelrun)
    except DataNotFoundError:
        LOGGER.error("Model run %s not found. Run 'smif list' to see available model runs.",
                     modelrun)
        exit(-1)

    LOGGER.info("Running %s", model_run_config['name'])
    LOGGER.debug("Model Run: %s", model_run_config)
    sos_model_config = handler.read_sos_model(model_run_config['sos_model'])

    sector_model_objects = []
    for sector_model in sos_model_config['sector_models']:
        sector_model_config = handler.read_sector_model(sector_model)

        absolute_path = os.path.join(directory,
                                     sector_model_config['path'])
        sector_model_config['path'] = absolute_path

        intervention_files = sector_model_config['interventions']
        intervention_list = []
        for intervention_file in intervention_files:
            interventions = handler.read_interventions(intervention_file)
            intervention_list.extend(interventions)
        sector_model_config['interventions'] = intervention_list

        initial_condition_files = sector_model_config['initial_conditions']
        initial_condition_list = []
        for initial_condition_file in initial_condition_files:
            initial_conditions = handler.read_initial_conditions(initial_condition_file)
            initial_condition_list.extend(initial_conditions)
        sector_model_config['initial_conditions'] = initial_condition_list

        sector_model_builder = SectorModelBuilder(sector_model_config['name'])
        # LOGGER.debug("Sector model config: %s", sector_model_config)
        sector_model_builder.construct(sector_model_config,
                                       model_run_config['timesteps'])
        sector_model_object = sector_model_builder.finish()

        sector_model_objects.append(sector_model_object)
        LOGGER.debug("Model inputs: %s", sector_model_object.inputs.names)

    LOGGER.debug("Sector models: %s", sector_model_objects)
    sos_model_config['sector_models'] = sector_model_objects

    scenario_objects = []
    for scenario_set, scenario_name in model_run_config['scenarios'].items():
        scenario_definition = handler.read_scenario_definition(scenario_name)
        LOGGER.debug("Scenario definition: %s", scenario_definition)

        scenario_model_builder = ScenarioModelBuilder(scenario_set)
        scenario_model_builder.construct(scenario_definition)
        scenario_objects.append(scenario_model_builder.finish())

    LOGGER.debug("Scenario models: %s", [model.name for model in scenario_objects])
    sos_model_config['scenario_sets'] = scenario_objects

    strategies = []
    for strategy in model_run_config['strategies']:
        if strategy['strategy'] == 'pre-specified-planning':
            interventions = handler.read_strategies(strategy['filename'])
            del strategy['filename']
            strategy['interventions'] = interventions
            LOGGER.debug("Added %s pre-specified planning interventions to %s",
                         len(interventions), strategy['model_name'])
        strategies.append(strategy)
    sos_model_config['strategies'] = strategies

    sos_model_builder = SosModelBuilder()
    sos_model_builder.construct(sos_model_config)
    sos_model_object = sos_model_builder.finish()

    LOGGER.debug("Model list: %s", list(sos_model_object.models.keys()))

    model_run_config['sos_model'] = sos_model_object
    narrative_objects = get_narratives(handler,
                                       model_run_config['narratives'])
    model_run_config['narratives'] = narrative_objects

    return model_run_config


def get_narratives(handler, narrative_config):
    """Load the narrative data from the sos model run configuration

    Arguments
    ---------
    handler: :class:`smif.data_layer.DataInterface`
    narrative_config: dict
        A dict with keys as narrative_set names and values as narrative names

    Returns
    -------
    list
        A list of :class:`smif.parameter.Narrative` objects populated with
        data

    """
    narrative_objects = []
    for narrative_set, narrative_names in narrative_config.items():
        LOGGER.info("Loading narrative data for narrative set '%s'",
                    narrative_set)
        for narrative_name in narrative_names:
            LOGGER.debug("Adding narrative entry '%s'", narrative_name)
            definition = handler.read_narrative_definition(narrative_name)
            narrative = Narrative(
                narrative_name,
                definition['description'],
                narrative_set
            )
            narrative.data = handler.read_narrative_data(narrative_name)
            narrative_objects.append(narrative)
    return narrative_objects


def build_model_run(model_run_config):
    """Builds the model run

    Arguments
    ---------
    model_run_config: dict
        A valid model run configuration dict with objects

    Returns
    -------
    `smif.modelrun.ModelRun`
    """
    try:
        builder = ModelRunBuilder()
        builder.construct(model_run_config)
        modelrun = builder.finish()
    except AssertionError as error:
        err_type, err_value, err_traceback = sys.exc_info()
        traceback.print_exception(err_type, err_value, err_traceback)
        err_msg = str(error)
        if err_msg:
            LOGGER.error("An AssertionError occurred (%s) see details above.", err_msg)
        else:
            LOGGER.error("An AssertionError occurred, see details above.")
        exit(-1)

    return modelrun



"""Memory-backed data interface
"""
from smif.data_layer.data_interface import DataInterface


class MemoryInterface(DataInterface):
    """ Read and write interface to main memory
    """
    def __init__(self):
        self._sos_model_runs = {}
        self._sos_models = {}
        self._sector_models = {}
        self._units = {}
        self._regions = {}
        self._intervals = {}
        self._scenario_sets = {}
        self._scenarios = {}
        self._narrative_sets = {}
        self._narratives = {}
        self._results = {}
        self._coefficients = {}
        self._strategies = {}

    def read_strategies(self):
        return self._strategies.items()

    def read_units_file_name(self):
        return self._units.values()

    def read_sos_model_runs(self):
        return self._sos_model_runs.values()

    def read_sos_model_run(self, sos_model_run_name):
        return self._sos_model_runs[sos_model_run_name]

    def write_sos_model_run(self, sos_model_run):
        self._sos_model_runs[sos_model_run['name']] = sos_model_run

    def update_sos_model_run(self, sos_model_run_name, sos_model_run):
        self._sos_model_runs[sos_model_run_name] = sos_model_run

    def delete_sos_model_run(self, sos_model_run):
        del self._sos_model_runs[sos_model_run['name']]

    def read_sos_models(self):
        return self._sos_models.values()

    def write_sos_model(self, sos_model):
        self._sos_models[sos_model['name']] = sos_model

    def update_sos_model(self, sos_model_name, sos_model):
        self._sos_models[sos_model_name] = sos_model

    def read_sector_models(self):
        return self._sector_models.values()

    def read_sector_model(self, sector_model_name):
        return self._sector_models[sector_model_name]

    def write_sector_model(self, sector_model):
        self._sector_models[sector_model['name']] = sector_model

    def update_sector_model(self, sector_model_name, sector_model):
        self._sector_models[sector_model_name] = sector_model

    def read_region_definitions(self):
        return self._regions.values()

    def read_region_definition_data(self, region_name):
        return self._regions[region_name]

    def write_region_definition(self, region):
        self._regions[region['name']] = region

    def update_region_definition(self, region):
        self._regions[region['name']] = region

    def read_interval_definitions(self):
        return self._intervals.values()

    def read_interval_definition_data(self, interval_name):
        return self._intervals[interval_name]

    def write_interval_definition(self, interval):
        self._intervals[interval['name']] = interval

    def update_interval_definition(self, interval):
        self._intervals[interval['name']] = interval

    def read_scenario_sets(self):
        return self._scenario_sets.values()

    def read_scenario_set(self, scenario_set_name):
        return self._scenario_sets[scenario_set_name]

    def write_scenario_set(self, scenario_set):
        self._scenario_sets[scenario_set['name']] = scenario_set

    def update_scenario_set(self, scenario_set):
        self._scenario_sets[scenario_set['name']] = scenario_set

    def read_scenario_data(self, scenario_name, parameter_name,
                           spatial_resolution, temporal_resolution, timestep):
        return self._scenarios[(
                scenario_name, parameter_name, spatial_resolution,
                temporal_resolution, timestep
            )]

    def write_scenario_data(self, scenario_name, parameter_name, data,
                            spatial_resolution, temporal_resolution, timestep):
        self._scenarios[(
            scenario_name, parameter_name, spatial_resolution,
            temporal_resolution, timestep
        )] = data

    def write_scenario(self, scenario):
        self._scenarios[scenario['name']] = scenario

    def update_scenario(self, scenario):
        self._scenarios[scenario['name']] = scenario

    def read_narrative_sets(self):
        return self._narrative_sets.values()

    def read_narrative_set(self, narrative_set_name):
        return self._narrative_sets[narrative_set_name]

    def write_narrative_set(self, narrative_set):
        self._narrative_sets[narrative_set['name']] = narrative_set

    def update_narrative_set(self, narrative_set):
        self._narrative_sets[narrative_set['name']] = narrative_set

    def read_narrative_data(self, narrative_name):
        return self._narratives[narrative_name]

    def write_narrative(self, narrative):
        self._narratives[narrative['name']] = narrative

    def update_narrative(self, narrative):
        self._narratives[narrative['name']] = narrative

    def read_results(self, modelrun_name, model_name, output_name, spatial_resolution,
                     temporal_resolution, timestep=None, modelset_iteration=None,
                     decision_iteration=None):
        return self._results[
            (
                modelrun_name, model_name, output_name, spatial_resolution,
                temporal_resolution, timestep, modelset_iteration,
                decision_iteration
            )]

    def write_results(self, modelrun_name, model_name, output_name, data, spatial_resolution,
                      temporal_resolution, timestep=None, modelset_iteration=None,
                      decision_iteration=None):
        self._results[
            (
                modelrun_name, model_name, output_name, spatial_resolution,
                temporal_resolution, timestep, modelset_iteration,
                decision_iteration
            )] = data

    def read_coefficients(self, source_name, destination_name):
        return self._coefficients[(source_name, destination_name)]

    def write_coefficients(self, source_name, destination_name, data):
        self._coefficients[(source_name, destination_name)] = data




from unittest.mock import Mock, patch

from smif.controller.scheduler import Scheduler


@patch('smif.controller.scheduler.subprocess.Popen')
def test_single_modelrun(mock_popen):
    my_scheduler = Scheduler()
    my_scheduler.add('my_model_run', {
        'directory': 'mock/dir',
        'verbosity': 0, 
        'warm_start': False, 
        'output_format': 'local_csv'
    })

    mock_popen.assert_called_with(
        'exec smif  run my_model_run -d mock/dir -i local_csv',
        shell=True,
        stderr=-2, stdout=-1
    )


def test_status_modelrun_never_added():
    my_scheduler = Scheduler()
    status = my_scheduler.get_status('my_model_run')
    assert status['status'] == 'unstarted'


@patch('smif.controller.scheduler.subprocess.Popen')
def test_status_model_started(mock_popen):

    attrs = {
        'poll.return_value': None,
        'communicate.return_value': (
            "this is a stdout".encode('utf-8'),
        ),
        'returncode': None
    }
    process_mock = Mock(**attrs)
    mock_popen.return_value = process_mock

    my_scheduler = Scheduler()
    my_scheduler.add('my_model_run', {
        'directory': 'mock/dir',
        'verbosity': 0, 
        'warm_start': False, 
        'output_format': 'local_csv'
    })
    my_scheduler.lock = True
    status = my_scheduler.get_status('my_model_run')
    assert status['status'] == 'running'


@patch('smif.controller.scheduler.subprocess.Popen')
def test_status_model_done(mock_popen):

    attrs = {
        'poll.return_value': 0,
        'communicate.return_value': (
            "this is a stdout".encode('utf-8')
        )
    }
    process_mock = Mock(**attrs)
    mock_popen.return_value = process_mock

    my_scheduler = Scheduler()
    my_scheduler.add('my_model_run', {
        'directory': 'mock/dir',
        'verbosity': 0, 
        'warm_start': False, 
        'output_format': 'local_csv'
    })
    my_scheduler.lock = True
    response = my_scheduler.get_status('my_model_run')

    assert response['status'] == 'done'


@patch('smif.controller.scheduler.subprocess.Popen')
def test_status_model_failed(mock_popen):

    attrs = {
        'poll.return_value': 1,
        'communicate.return_value': (
            "this is a stdout".encode('utf-8'),
        )
    }
    process_mock = Mock(**attrs)
    mock_popen.return_value = process_mock

    my_scheduler = Scheduler()
    my_scheduler.add('my_model_run', {
        'directory': 'mock/dir',
        'verbosity': 0, 
        'warm_start': False, 
        'output_format': 
        'local_csv'
    })
    my_scheduler.lock = True
    response = my_scheduler.get_status('my_model_run')

    assert response['status'] == 'failed'


@patch('smif.controller.scheduler.subprocess.Popen')
def test_status_model_stopped(mock_popen):

    attrs = {
        'poll.return_value': None,
        'communicate.return_value': (
            "this is a stdout".encode('utf-8'),
        )
    }
    process_mock = Mock(**attrs)
    mock_popen.return_value = process_mock

    my_scheduler = Scheduler()
    my_scheduler.add('my_model_run', {
        'directory': 'mock/dir', 
        'verbosity': 0, 
        'warm_start': False, 
        'output_format': 
        'local_csv'
    })
    my_scheduler.lock = True
    my_scheduler.kill('my_model_run')
    response = my_scheduler.get_status('my_model_run')

    assert response['status'] == 'stopped'



"""Some Metadata for model parameters, understood as single values with a
default and an expected range.

"""
from smif.parameters.narrative import Narrative
from smif.parameters.parameters import Parameter, ParameterList

__all__ = ['Parameter', 'ParameterList', 'Narrative']



import numpy as np
from numpy.testing import assert_equal
from pytest import raises
from smif.convert.interval import Interval, IntervalSet, TimeIntervalRegister


class TestInterval:

    def test_empty_interval_list(self):

        with raises(ValueError):
            Interval('test', [])

    def test_equality(self):
        a = Interval('test', ('PT0H', 'PT1H'))
        b = Interval('test', ('PT0H', 'PT1H'))
        c = Interval('another_test', ('PT0H', 'PT1H'))
        d = Interval('another_test', ('PT1H', 'PT2H'))

        assert a == b
        assert a != c
        assert b != c
        assert c != d

    def test_invalid_set_interval(self):
        interval = Interval('test', ('PT0H', 'PT1H'))
        with raises(ValueError) as excinfo:
            interval.interval = None
        msg = "A time interval must add either a single tuple or a list of tuples"
        assert msg in str(excinfo)

    def test_empty_interval_tuple(self):
        with raises(ValueError):
            Interval('test', ())

    def test_empty_interval_list_tuple(self):
        with raises(ValueError):
            Interval('test', [()])

    def test_load_interval(self):

        interval = Interval('test', ('PT0H', 'PT1H'))

        assert interval.name == 'test'
        assert interval.start == 'PT0H'
        assert interval.end == 'PT1H'

    def test_load_multiple_interval(self):

        interval = Interval('test', [('PT0H', 'PT1H'), ('PT1H', 'PT2H')])

        assert interval.name == 'test'
        assert interval.start == ['PT0H', 'PT1H']
        assert interval.end == ['PT1H', 'PT2H']

    def test_add_list_of_interval(self):

        interval = Interval('test', ('PT0H', 'PT1H'))
        interval.interval = [('PT1H', 'PT2H')]

        assert interval.name == 'test'
        assert interval.start == ['PT0H', 'PT1H']
        assert interval.end == ['PT1H', 'PT2H']

    def test_raise_error_load_illegal(self):
        with raises(ValueError):
            Interval('test', 'start', 'end')

    def test_convert_hour_to_hours(self):

        interval = Interval('test', ('PT0H', 'PT1H'))
        actual = interval._convert_to_hours(interval.start)
        expected = 0
        assert actual == expected

        actual = interval._convert_to_hours(interval.end)
        expected = 1
        assert actual == expected

    def test_convert_month_to_hours(self):

        interval = Interval('test', ('P1M', 'P2M'))
        actual = interval._convert_to_hours(interval.start)
        expected = 744
        assert actual == expected

        actual = interval._convert_to_hours(interval.end)
        expected = 1416
        assert actual == expected

    def test_convert_week_to_hours(self):

        interval = Interval('test', ('P2D', 'P3D'))
        actual = interval._convert_to_hours(interval.start)
        expected = 48
        assert actual == expected

        actual = interval._convert_to_hours(interval.end)
        expected = 72
        assert actual == expected

    def test_to_hours_zero(self):

        interval = Interval('test', ('PT0H', 'PT1H'))
        actual = interval.bounds

        assert actual == [(0, 1)]

    def test_to_hours_month(self):

        interval = Interval('test', ('P2M', 'P3M'))
        actual = interval.bounds

        assert actual == [(1416, 2160)]

    def test_to_hours_list(self):

        interval = Interval('test', [('PT0H', 'PT1H'),
                                     ('PT2H', 'PT3H'),
                                     ('PT5H', 'PT7H')])
        actual = interval.bounds

        assert actual == [(0, 1), (2, 3), (5, 7)]

    def test_str_one_interval(self):
        interval = Interval('test', ('P2M', 'P3M'))
        actual = str(interval)
        assert actual == \
            "Interval 'test' maps to:\n  hour 1416 to hour 2160\n"

    def test_str_multiple_intervals(self):
        interval = Interval('test', [('PT0H', 'PT1H'),
                                     ('PT2H', 'PT3H'),
                                     ('PT5H', 'PT7H')])
        actual = str(interval)
        assert actual == \
            "Interval 'test' maps to:\n  hour 0 to hour 1\n  hour 2 to hour 3\n  "\
            "hour 5 to hour 7\n"

    def test_repr(self):
        interval = Interval('test', ('P2M', 'P3M'), 2011)
        actual = repr(interval)
        assert actual == "Interval('test', [('P2M', 'P3M')], base_year=2011)"

    def test_load_remap_timeslices(self):
        interval = Interval('1', [('P2M', 'P3M'),
                                  ('P3M', 'P4M'),
                                  ('P4M', 'P5M')])
        actual = interval.bounds
        assert actual == [(1416, 2160), (2160, 2880), (2880, 3624)]


class TestIntervalSet:

    def test_get_names(self, months):

        expected_names = \
            ['jan', 'feb', 'mar', 'apr', 'may', 'jun',
             'jul', 'aug', 'sep', 'oct', 'nov', 'dec']
        interval_set = IntervalSet('months', months)
        actual_names = interval_set.get_entry_names()
        assert expected_names == actual_names

    def test_attributes(self, months):
        interval_set = IntervalSet('months', months)
        interval_set.description = 'a descriptions'

        actual = interval_set.as_dict()
        expected = {'name': 'months',
                    'description': 'a descriptions'}
        assert actual == expected


class TestIntervalRegister():

    def test_interval_loads(self):
        """Pass a time-interval definition into the register

        """
        data = [('1_1', [('PT0H', 'PT1H')])]

        register = TimeIntervalRegister()
        register.register(IntervalSet('energy_supply_hourly', data))
        assert register.names == ['energy_supply_hourly']

        actual = register.get_entry('energy_supply_hourly')

        element = Interval('1_1', ('PT0H', 'PT1H'), base_year=2010)
        expected = []
        expected.append(element)

        assert actual.data == expected

    def test_interval_load_duplicate_name_raises(self, months):
        """Tests that error is raised if a duplicate name is used
        for an interval set
        """
        register = TimeIntervalRegister()
        register.register(IntervalSet('months', months))
        with raises(ValueError):
            register.register(IntervalSet('months', months))

    def test_months_load(self, months):
        """Pass a monthly time-interval definition into the register

        """
        register = TimeIntervalRegister()
        register.register(IntervalSet('months', months))

        actual = register.get_entry('months')

        expected = [Interval('jan', ('P0M', 'P1M')),
                    Interval('feb', ('P1M', 'P2M')),
                    Interval('mar', ('P2M', 'P3M')),
                    Interval('apr', ('P3M', 'P4M')),
                    Interval('may', ('P4M', 'P5M')),
                    Interval('jun', ('P5M', 'P6M')),
                    Interval('jul', ('P6M', 'P7M')),
                    Interval('aug', ('P7M', 'P8M')),
                    Interval('sep', ('P8M', 'P9M')),
                    Interval('oct', ('P9M', 'P10M')),
                    Interval('nov', ('P10M', 'P11M')),
                    Interval('dec', ('P11M', 'P12M'))]

        for idx, interval in enumerate(expected):
            assert actual.data[idx] == interval

    def test_remap_interval_load(self, remap_months):
        register = TimeIntervalRegister()

        intervals = IntervalSet('remap_months', remap_months)

        register.register(intervals)

        actual = register.get_entry('remap_months')

        assert actual == intervals

    def test_conversion_with_different_coverage_fails(self, one_year, one_day, caplog):

        register = TimeIntervalRegister()
        register.register(IntervalSet("one_year", one_year))
        register.register(IntervalSet("one_day", one_day))

        data = np.array([[1]])
        register.convert(data, 'one_year', 'one_day')

        expected = "Coverage for 'one_year' is 8760 and does not match " \
                   "coverage for 'one_day' which is 24"

        assert expected in caplog.text


class TestTimeRegisterCoefficients:

    def test_coeff(self, months, seasons, month_to_season_coefficients):

        register = TimeIntervalRegister()
        register.register(IntervalSet('months', months))
        register.register(IntervalSet('seasons', seasons))

        actual = register.get_coefficients('months', 'seasons')
        expected = month_to_season_coefficients
        assert np.allclose(actual, expected, rtol=1e-05, atol=1e-08)

    def test_time_only_conversion(self, months, seasons):

        register = TimeIntervalRegister()
        register.register(IntervalSet('months', months))
        register.register(IntervalSet('seasons', seasons))
        data = np.array([[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]])
        actual = register.convert(data, 'months', 'seasons')
        expected = np.array([[3, 3, 3, 3]])
        np.testing.assert_array_equal(actual, expected)

    def test_time_only_conversion_disagg(self, months, seasons):

        register = TimeIntervalRegister()
        register.register(IntervalSet('months', months))
        register.register(IntervalSet('seasons', seasons))
        data = np.array([[3, 3, 3, 3]])
        actual = register.convert(data, 'seasons', 'months')
        expected = np.array([[1.033333, 0.933333, 1.01087, 0.978261,
                              1.01087, 0.978261, 1.01087, 1.01087,
                              0.989011, 1.021978, 0.989011, 1.033333]])
        np.testing.assert_allclose(actual, expected, rtol=1e-3)


class TestTimeRegisterConversion:

    def test_raises_error_on_no_definition(self):

        register = TimeIntervalRegister()
        with raises(ValueError):
            register.get_entry('blobby')

    def test_aggregate_from_month_to_seasons(self,
                                             months,
                                             seasons,
                                             monthly_data,
                                             monthly_data_as_seasons):
        register = TimeIntervalRegister()
        register.register(IntervalSet('months', months))
        register.register(IntervalSet('seasons', seasons))

        actual = register.convert(monthly_data, 'months', 'seasons')
        expected = monthly_data_as_seasons
        assert np.allclose(actual, expected, rtol=1e-05, atol=1e-08)

    def test_agggregate_from_hour_to_day(self, twenty_four_hours, one_day):

        data = np.ones((1, 24))

        register = TimeIntervalRegister()
        register.register(IntervalSet('hourly_day', twenty_four_hours))
        register.register(IntervalSet('one_day', one_day))

        actual = register.convert(data, 'hourly_day', 'one_day')
        expected = np.array([24])

        assert np.allclose(actual, expected, rtol=1e-05, atol=1e-08)


class TestValidation:

    def test_validate_get_hourly_array(self, remap_months):
        intervals = IntervalSet('remap_months', remap_months)

        actual = intervals._get_hourly_array()
        expected = np.ones(8760, dtype=np.int)
        assert_equal(actual, expected)

    def test_validate_intervals_passes(self, remap_months):

        register = TimeIntervalRegister()
        register.register(IntervalSet('remap_months', remap_months))

    def test_validate_intervals_fails(self, remap_months):
        data = remap_months
        data.append(('5', [('PT0H', 'PT1H')]))
        with raises(ValueError) as excinfo:
            IntervalSet('remap_months', data)
        assert "Duplicate entry for hour 0 in interval set remap_months." in str(excinfo.value)

    def test_time_interval_start_before_end(self):
        with raises(ValueError) as excinfo:
            Interval('backwards', ('P1Y', 'P3M'))
        assert "A time interval must not end before it starts" in str(excinfo)

        interval = Interval('starts_ok', ('P0Y', 'P1M'))
        with raises(ValueError) as excinfo:
            interval.interval = ('P2M', 'P1M')
        assert "A time interval must not end before it starts" in str(excinfo)


class TestIntersection:

    def test_intersection(self, months, seasons):

        month_set = IntervalSet('months', months)
        season_set = IntervalSet('seasons', seasons)

        actual = month_set.intersection(season_set.data[0])
        expected = [0, 1, 11]
        assert actual == expected

    def test_intersection_seasons(self, months, seasons):

        month_set = IntervalSet('months', months)
        season_set = IntervalSet('seasons', seasons)
        actual = season_set.intersection(month_set.data[0])
        expected = [0]
        assert actual == expected


class TestBounds:

    def test_bounds_checker(self, months):

        bounds = [(0, 1), (1, 2)]

        expected = True

        time_set = IntervalSet('months', months)
        actual = time_set.check_interval_bounds_equal(bounds)

        assert actual == expected

    def test_bounds_checker_fails(self, months):

        bounds = [(0, 8759), (8759, 8760)]

        expected = False

        time_set = IntervalSet('months', months)
        actual = time_set.check_interval_bounds_equal(bounds)

        assert actual == expected


class TestProportions:

    def test_get_proportion_month_in_season(self, months, seasons):

        month_set = IntervalSet('months', months)
        season_set = IntervalSet('seasons', seasons)

        winter = season_set.data[0]

        actual = month_set.get_proportion(0, winter)
        expected = 1
        assert actual == expected

    def test_get_proportion_season_in_month(self, months, seasons):

        month_set = IntervalSet('months', months)
        season_set = IntervalSet('seasons', seasons)

        january = month_set.data[0]

        actual = season_set.get_proportion(0, january)
        expected = 31 * 1 / (31+31+28)
        np.testing.assert_allclose(actual, expected)

    def test_remap_disagg_proportions(self, months, remap_months):
        """Find proportion of January in cold_month, split over year
        """

        month_set = IntervalSet('months', months)
        remap_set = IntervalSet('remap_months', remap_months)

        to_interval = remap_set.data[0]  # cold month

        actual = month_set.get_proportion(0, to_interval)
        expected = 0.33333
        np.testing.assert_allclose(actual, expected, rtol=1e-3)

    def test_remap_agg_proportions(self, months, remap_months):

        month_set = IntervalSet('months', months)
        remap_set = IntervalSet('remap_months', remap_months)

        to_interval = month_set.data[0]  # january

        actual = remap_set.get_proportion(0, to_interval)
        expected = 1.0333333333333
        np.testing.assert_allclose(actual, expected, rtol=1e-3)


class TestRemapConvert:

    def test_remap_coefficients(self, months, remap_months):
        """Twelve months are remapped (averaged) into four
        representative months
        """
        register = TimeIntervalRegister()
        register.register(IntervalSet('months', months))
        register.register(IntervalSet('remap_months', remap_months))
        actual = register.get_coefficients('months', 'remap_months')

        expected = np.array([[0.333333, 0, 0, 0],
                             [0.333333, 0, 0, 0],
                             [0, 0.333333, 0, 0],
                             [0, 0.333333, 0, 0],
                             [0, 0.333333, 0, 0],
                             [0, 0, 0.333333, 0],
                             [0, 0, 0.333333, 0],
                             [0, 0, 0.333333, 0],
                             [0, 0, 0, 0.333333],
                             [0, 0, 0, 0.333333],
                             [0, 0, 0, 0.333333],
                             [0.333333, 0, 0, 0]]
                            )

        np.testing.assert_allclose(actual, expected, rtol=1e-3)

    def test_resample_mapped_months_to_months(self, months, remap_months):
        """Converts from remapped month data (where one average month is used
        for each season) back to months
        """
        register = TimeIntervalRegister()
        register.register(IntervalSet('remap_months', remap_months))
        register.register(IntervalSet('months', months))

        data = np.array([[1, 1, 1, 1]])
        actual = register.convert(data,
                                  'remap_months',
                                  'months')
        expected = np.array([[1.033333, 0.933333, 1.01087, 0.978261, 1.01087,
                              0.978261, 1.01087, 1.01087, 0.989011, 1.021978,
                              0.989011, 1.033333]])

        np.testing.assert_allclose(actual, expected, rtol=1e-3)

    def test_remap_months_to_mapped_months(self,
                                           months,
                                           monthly_data,
                                           remap_months,
                                           remap_month_data):
        register = TimeIntervalRegister()
        register.register(IntervalSet('months', months))
        register.register(IntervalSet('remap_months', remap_months))

        actual = register.convert(monthly_data, 'months', 'remap_months')
        expected = remap_month_data

        np.testing.assert_allclose(actual, expected, rtol=1e-05, atol=1e-08)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Water Supply model
- WaterSupplySectorModel implements SectorModel
- wraps ExampleWaterSupplySimulationModel
- instantiate a model instance

"""

import logging

import numpy as np
from smif import StateData
from smif.model.sector_model import SectorModel


class WaterSupplySectorModel(SectorModel):
    """Example of a class implementing the SectorModel interface,
    using one of the toy water models below to simulate the water supply
    system.
    """

    def initialise(self, initial_conditions):
        """Set up system here
        """
        pass

    def simulate(self, data):
        """

        Arguments
        =========
        data
            - scenario data, e.g. expected level of rainfall
            - decisions
                - asset build instructions, demand-side interventions to apply
            - state
                - existing system/network (unless implementation means maintaining
                system in sector model)
                - system state, e.g. reservoir level at year start
        """

        self.logger.debug("Initial conditions: %s", self._initial_state)

        state = self._initial_state[0]

        # unpack inputs
        per_capita_water_demand = \
            data.get_parameter('per_capita_water_demand')  # liter/person

        population = data.get_data('population')  # people

        water_demand = (population * per_capita_water_demand) + \
            data.get_data('water_demand')  # liter

        raininess = sum(data.get_data('raininess'))  # megaliters

        reservoir_level = state.data['current_level']['value'] \
            - (1e-6 * water_demand.sum())  # megaliters

        self.logger.debug("Parameters:\n  Population: %s\n  Raininess: %s\n "
                          "Reservoir level: %s\n  ", population.sum(),
                          raininess.sum(), reservoir_level)

        # unpack assets
        number_of_treatment_plants = 2

        self.logger.debug(state)
        self.logger.debug(data.get_parameters())

        # simulate (wrapping toy model)
        instance = ExampleWaterSupplySimulationModel()
        instance.raininess = raininess
        instance.number_of_treatment_plants = number_of_treatment_plants
        instance.reservoir_level = reservoir_level

        water, cost = instance.run()
        data.set_results('water', np.ones((3, 1)) * water / 3)
        data.set_results("cost", np.ones((3, 1)) * cost / 3)
        data.set_results("energy_demand", np.ones((3, 1)) * 3)

        state = StateData('Kielder Water', {
            'current_level': {'value': instance.reservoir_level}
        })

    def extract_obj(self, results):
        return results['cost'].sum()


class ExampleWaterSupplySimulationModel(object):
    """An example simulation model used for testing purposes

    Parameters
    ==========
    raininess : int
        The amount of rain produced in each simulation
    number_of_treatment_plants : int
        The amount of water is a function of the number of treatment plants and
        the amount of raininess
    """
    def __init__(self,
                 raininess=None,
                 number_of_treatment_plants=None,
                 reservoir_level=None):
        self.raininess = raininess
        self.number_of_treatment_plants = number_of_treatment_plants
        self.reservoir_level = reservoir_level

    def run(self):
        """Runs the water supply model

        Only 1 unit of water is produced per treatment plant,
        no matter how rainy.

        Each treatment plant costs 1.0 unit.
        """
        logger = logging.getLogger(__name__)

        logger.debug("There are %s plants", self.number_of_treatment_plants)
        logger.debug("It is %s rainy", self.raininess)

        logger.debug("Reservoir level was %s", self.reservoir_level)
        self.reservoir_level += self.raininess

        water = min(self.number_of_treatment_plants, self.reservoir_level)
        logger.debug("The system produces %s water", water)

        self.reservoir_level -= water
        logger.debug("Reservoir level now %s", self.reservoir_level)

        cost = 1.264 * self.number_of_treatment_plants
        logger.debug("The system costs £%s", cost)

        return water, cost


if __name__ == '__main__':
    """Run core model if this script is run from the command line
    """
    CORE_MODEL = ExampleWaterSupplySimulationModel(1, 1, 2)
    CORE_MODEL.run()



"""Provide method for creating the smif app

For example, build the front end::
>>> cd smif/app/
>>> npm run build

Or to rebuild js/css on change::
>>> npm run watch

Then run the server app in debug mode with environment variables
>>> FLASK_DEBUG=1 smif -v app -d src/smif/sample_project/

Or if backend debug mode is not needed, just use the smif CLI::
>>> smif app -d ../sample_project

API commands can be operated without the front end

GET commands can be performed by going to the requested url::
>>> firefox http://localhost:5000/api/v1/sos_model_runs

POST/PUT commands on configurations can be performed by using the curl utilities under linux::
>>> curl -d '{
...     "name": "scenario_set",
...     "description": "my description",
...     "facets": []
... }' -H "Content-Type: application/json" -X POST http://localhost:5000/api/v1/scenario_sets/

Actions are triggered by sending an empty POST::
>>> curl -d '{}' http://localhost:5000/api/v1/sos_model_runs/20170918_energy_water/start
"""

# import classes for access like ::
#         from smif.http_api import create_app
from smif.http_api.app import create_app

# Define what should be imported as * ::
#         from smif.http_api import *
__all__ = ['create_app']



"""Handles conversion between units used in the `SosModel`

First implementation delegates to pint.
"""
import logging

from pint import DimensionalityError, UndefinedUnitError, UnitRegistry
from smif.convert.register import Register


class UnitRegister(Register):

    def __init__(self):
        self._register = UnitRegistry(on_redefinition='raise')
        self.LOGGER = logging.getLogger()
        self.axis = None

    @property
    def names(self):
        return list(self._register.__dict__['_units'].keys())

    def register(self, unit_file):
        """Load unit definitions into the registry
        """
        self._register.load_definitions(unit_file)
        self.LOGGER.info("Finished registering user defined units")

        with open(unit_file, 'r') as readonlyfile:
            self.LOGGER.info("Imported user units:")
            for line in readonlyfile:
                self.LOGGER.info("    %s", line.split('=')[0])

    def get_entry(self, name):
        pass

    def get_coefficients(self, source, destination):
        return self.convert_old(1, source, destination)

    def convert_old(self, data, from_unit, to_unit):
        """Convert the data from one unit to another unit

        Parameters
        ----------
        data: numpy.ndarray
            An array of values with dimensions regions x intervals
        from_unit: str
            The name of the unit of the data
        to_unit: str
            The name of the required unit

        Returns
        -------
        numpy.ndarray
            An array of data with dimensions regions x intervals

        Raises
        ------
        ValueError
            If the units are not in the unit register or conversion is not possible
        """
        try:
            Q_ = self._register.Quantity(data, from_unit)
        except UndefinedUnitError:
            raise ValueError('Cannot convert from undefined unit ' + from_unit)

        try:
            result = Q_.to(to_unit).magnitude
        except UndefinedUnitError:
            raise ValueError('Cannot convert to undefined unit ' + to_unit)
        except DimensionalityError:
            raise ValueError('Cannot convert from ' + from_unit + ' to ' + to_unit)

        return result

    def parse_unit(self, unit_string):
        """Parse a unit string (abbreviation or full) into a Unit object

        Parameters
        ----------
        unit : str

        Returns
        -------
        quantity : :class:`pint.Unit`
        """
        try:
            unit = self._register.parse_units(unit_string)
        except UndefinedUnitError:
            self.LOGGER.warning("Unrecognised unit: %s", unit_string)
            unit = None
        return unit


__REGISTER = UnitRegister()


def get_register():
    """Returns a reference to the unit registry
    """
    return __REGISTER
